# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os
import sys
from os.path import dirname, abspath

import pyomo.common.unittest as unittest

from pyomo.environ import (
    ConcreteModel,
    Var,
    Objective,
    RangeSet,
    Constraint,
    Reals,
    NonNegativeIntegers,
    NonNegativeReals,
    Integers,
    Binary,
    maximize,
    minimize,
)
from pyomo.opt import SolverFactory, TerminationCondition, SolverStatus
from pyomo.solvers.plugins.solvers.CBCplugin import CBCSHELL

cbc_available = SolverFactory('cbc', solver_io='lp').available(exception_flag=False)

data_dir = '{}/data'.format(dirname(abspath(__file__)))


@unittest.skipIf(not cbc_available, "The 'cbc' solver is not available")
@unittest.pytest.mark.solver("cbc")
class TestCBC(unittest.TestCase):
    """
    These tests are here to test the general functionality of the cbc solver when using the lp solverio, which will
    ensure that we have a consistent output from CBC for some simple problems
    """

    def setUp(self):
        self.stderr = sys.stderr
        sys.stderr = None
        self.model = ConcreteModel()
        # Do we need to pass in seeds to ensure consistent behaviour? options={'randomSeed: 42, 'randomCbcSeed': 42}
        self.opt = SolverFactory("cbc", solver_io="lp")

    def tearDown(self):
        sys.stderr = self.stderr

    def test_infeasible_lp(self):
        self.model.X = Var(within=Reals)
        self.model.C1 = Constraint(expr=self.model.X <= 1)
        self.model.C2 = Constraint(expr=self.model.X >= 2)
        self.model.Obj = Objective(expr=self.model.X, sense=minimize)

        results = self.opt.solve(self.model)

        self.assertEqual(minimize, results.problem.sense)
        self.assertEqual(
            TerminationCondition.infeasible, results.solver.termination_condition
        )
        self.assertEqual(
            'Model was proven to be infeasible.', results.solver.termination_message
        )
        self.assertEqual(SolverStatus.warning, results.solver.status)

    def test_unbounded_lp(self):
        self.model.Idx = RangeSet(2)
        self.model.X = Var(self.model.Idx, within=Reals)
        self.model.Obj = Objective(
            expr=self.model.X[1] + self.model.X[2], sense=maximize
        )

        results = self.opt.solve(self.model)

        self.assertEqual(maximize, results.problem.sense)
        self.assertEqual(
            TerminationCondition.unbounded, results.solver.termination_condition
        )
        self.assertEqual(
            'Model was proven to be unbounded.', results.solver.termination_message
        )
        self.assertEqual(SolverStatus.warning, results.solver.status)

    def test_optimal_lp(self):
        self.model.X = Var(within=NonNegativeReals)
        self.model.Obj = Objective(expr=self.model.X, sense=minimize)

        results = self.opt.solve(self.model)

        self.assertEqual(0.0, results.problem.lower_bound)
        self.assertEqual(0.0, results.problem.upper_bound)
        self.assertEqual(minimize, results.problem.sense)
        self.assertEqual(
            TerminationCondition.optimal, results.solver.termination_condition
        )
        self.assertEqual(
            'Model was solved to optimality (subject to tolerances), and an optimal solution is available.',
            results.solver.termination_message,
        )
        self.assertEqual(SolverStatus.ok, results.solver.status)

    def test_infeasible_mip(self):
        self.model.X = Var(within=NonNegativeIntegers)
        self.model.C1 = Constraint(expr=self.model.X <= 1)
        self.model.C2 = Constraint(expr=self.model.X >= 2)
        self.model.Obj = Objective(expr=self.model.X, sense=minimize)

        results = self.opt.solve(self.model)

        self.assertEqual(minimize, results.problem.sense)
        self.assertEqual(
            TerminationCondition.infeasible, results.solver.termination_condition
        )
        self.assertEqual(
            'Model was proven to be infeasible.', results.solver.termination_message
        )
        self.assertEqual(SolverStatus.warning, results.solver.status)

    def test_unbounded_mip(self):
        self.model.X = Var(within=Integers)
        self.model.Obj = Objective(expr=self.model.X, sense=minimize)

        results = self.opt.solve(self.model)

        self.assertEqual(minimize, results.problem.sense)
        self.assertEqual(
            TerminationCondition.unbounded, results.solver.termination_condition
        )
        self.assertEqual(
            'Model was proven to be unbounded.', results.solver.termination_message
        )
        self.assertEqual(SolverStatus.warning, results.solver.status)

    def test_optimal_mip(self):
        self.model.Idx = RangeSet(2)
        self.model.X = Var(self.model.Idx, within=NonNegativeIntegers)
        self.model.Y = Var(self.model.Idx, within=Binary)
        self.model.C1 = Constraint(expr=self.model.X[1] == self.model.X[2] + 1)
        self.model.Obj = Objective(
            expr=self.model.Y[1] + self.model.Y[2] - self.model.X[1], sense=maximize
        )

        results = self.opt.solve(self.model)

        self.assertEqual(1.0, results.problem.lower_bound)
        self.assertEqual(1.0, results.problem.upper_bound)
        self.assertEqual(results.problem.number_of_binary_variables, 2)
        self.assertEqual(results.problem.number_of_integer_variables, 4)
        self.assertEqual(maximize, results.problem.sense)
        self.assertEqual(
            TerminationCondition.optimal, results.solver.termination_condition
        )
        self.assertEqual(
            'Model was solved to optimality (subject to tolerances), and an optimal solution is available.',
            results.solver.termination_message,
        )
        self.assertEqual(SolverStatus.ok, results.solver.status)


class TestCBCUsingMock(unittest.TestCase):
    """
    These tests cover various abnormal exit conditions from CBC (notably various time or solution limits).
    In order to be able to reliably compare solutions on different platforms, we will compare against cached solutions.
    Effectively we are just testing how we parse the output files.
    The files were produced by using the following configuration:
    macOS Mojave Version 10.14.3
    CBC Version 2.9.9
    appdirs==1.4.3
    certifi==2018.11.29
    nose==1.3.7
    numpy==1.16.1
    ply==3.11
    Pyomo==5.6.1
    PyUtilib==5.6.5
    six==1.12.0


    n = 19
    np.random.seed(42)
    distance_matrix = np.random.rand(n, n)
    model = ConcreteModel()
    model.N = RangeSet(n)
    model.c = Param(model.N, model.N, initialize=lambda a_model, i, j: distance_matrix[i - 1][j - 1])
    model.x = Var(model.N, model.N, within=Binary)
    model.u = Var(model.N, within=NonNegativeReals)

    # Remove arcs that go to and from same node
    for n in model.N:
        model.x[n, n].fix(0)

    def obj_rule(a_model):
        return sum(a_model.c[i, j] * a_model.x[i, j] for i in a_model.N for j in a_model.N)

    model.obj = Objective(rule=obj_rule, sense=minimize)

    def only_leave_each_node_once_constraints(a_model, i):
        return sum(a_model.x[i, j] for j in a_model.N) == 1

    def only_arrive_at_each_node_once_constraints(a_model, j):
        return sum(a_model.x[i, j] for i in a_model.N) == 1

    def miller_tucker_zemlin_constraints(a_model, i, j):
        if i != j and i >= 2 and j >= 2:
            return a_model.u[i] - a_model.u[j] + a_model.x[i, j] * n <= n - 1
        return Constraint.NoConstraint

    model.con1 = Constraint(model.N, rule=only_leave_each_node_once_constraints)
    model.con2 = Constraint(model.N, rule=only_arrive_at_each_node_once_constraints)
    model.con3 = Constraint(model.N, model.N, rule=miller_tucker_zemlin_constraints)

    opt = SolverFactory('cbc', executable=_get_path_for_solver(), options={'randomCbcSeed': 42, 'randomSeed': 42})
    results = opt.solve(self.model, tee=True, **solver_kwargs)

    """

    def setUp(self):
        self.stderr = sys.stderr
        sys.stderr = None
        self.opt = SolverFactory("_mock_cbc", solver_io="lp")

    def tearDown(self):
        sys.stderr = self.stderr

    def test_optimal_mip(self):
        """
        solver_kwargs={}
        """
        lp_file = 'optimal.out.lp'
        results = self.opt.solve(os.path.join(data_dir, lp_file))

        self.assertEqual(1.20645976, results.problem.lower_bound)
        self.assertEqual(1.20645976, results.problem.upper_bound)
        self.assertEqual(SolverStatus.ok, results.solver.status)
        self.assertEqual(0.34, results.solver.system_time)
        self.assertEqual(0.72, results.solver.wallclock_time)
        self.assertEqual(
            TerminationCondition.optimal, results.solver.termination_condition
        )
        self.assertEqual(
            'Model was solved to optimality (subject to tolerances), and an optimal solution is available.',
            results.solver.termination_message,
        )
        self.assertEqual(
            results.solver.statistics.branch_and_bound.number_of_bounded_subproblems, 2
        )
        self.assertEqual(
            results.solver.statistics.branch_and_bound.number_of_created_subproblems, 2
        )
        self.assertEqual(results.solver.statistics.black_box.number_of_iterations, 625)

    def test_max_time_limit_mip(self):
        """
        solver_kwargs={'timelimit': 0.1}
        """
        lp_file = 'max_time_limit.out.lp'
        results = self.opt.solve(os.path.join(data_dir, lp_file))

        self.assertEqual(
            1.1084706, results.problem.lower_bound
        )  # Note that we ignore the lower bound given at the end
        self.assertEqual(1.35481947, results.problem.upper_bound)
        self.assertEqual(SolverStatus.aborted, results.solver.status)
        self.assertEqual(0.1, results.solver.system_time)
        self.assertEqual(0.11, results.solver.wallclock_time)
        self.assertEqual(
            TerminationCondition.maxTimeLimit, results.solver.termination_condition
        )
        self.assertEqual(
            'Optimization terminated because the time expended exceeded the value specified in the seconds parameter.',
            results.solver.termination_message,
        )
        self.assertEqual(
            results.solver.statistics.branch_and_bound.number_of_bounded_subproblems, 0
        )
        self.assertEqual(
            results.solver.statistics.branch_and_bound.number_of_created_subproblems, 0
        )
        self.assertEqual(results.solver.statistics.black_box.number_of_iterations, 82)

    def test_intermediate_non_integer_mip(self):
        """
        solver_kwargs={'timelimit': 0.0001}
        """
        lp_file = 'intermediate_non_integer.out.lp'
        results = self.opt.solve(os.path.join(data_dir, lp_file))

        self.assertEqual(0.92543678, results.problem.lower_bound)
        self.assertEqual(SolverStatus.aborted, results.solver.status)
        self.assertEqual(0.02, results.solver.system_time)
        self.assertEqual(0.02, results.solver.wallclock_time)
        self.assertEqual(
            TerminationCondition.intermediateNonInteger,
            results.solver.termination_condition,
        )
        self.assertEqual(
            'Optimization terminated because a limit was hit, however it had not found an integer solution yet.',
            results.solver.termination_message,
        )
        self.assertEqual(
            results.solver.statistics.branch_and_bound.number_of_bounded_subproblems, 0
        )
        self.assertEqual(
            results.solver.statistics.branch_and_bound.number_of_created_subproblems, 0
        )
        self.assertEqual(results.solver.statistics.black_box.number_of_iterations, 0)

    def test_max_solutions(self):
        """
        solver_kwargs={'options': {'maxSolutions': 1}}
        """
        lp_file = 'max_solutions.out.lp'
        results = self.opt.solve(os.path.join(data_dir, lp_file))

        self.assertEqual(0.92543678, results.problem.lower_bound)
        self.assertEqual(1.35481947, results.problem.upper_bound)
        self.assertEqual(SolverStatus.aborted, results.solver.status)
        self.assertEqual(0.03, results.solver.system_time)
        self.assertEqual(0.03, results.solver.wallclock_time)
        self.assertEqual(
            TerminationCondition.other, results.solver.termination_condition
        )
        self.assertEqual(
            'Optimization terminated because the number of solutions found reached the value specified in the '
            'maxSolutions parameter.',
            results.solver.termination_message,
        )
        self.assertEqual(
            results.solver.statistics.branch_and_bound.number_of_bounded_subproblems, 0
        )
        self.assertEqual(
            results.solver.statistics.branch_and_bound.number_of_created_subproblems, 0
        )
        self.assertEqual(results.solver.statistics.black_box.number_of_iterations, 0)

    def test_within_gap_tolerance(self):
        """
        solver_kwargs={'options': {'allowableGap': 1000000}}
        """
        lp_file = 'within_gap_tolerance.out.lp'
        results = self.opt.solve(os.path.join(data_dir, lp_file))

        self.assertEqual(0.925437, results.problem.lower_bound)
        self.assertEqual(1.35481947, results.problem.upper_bound)
        self.assertEqual(SolverStatus.ok, results.solver.status)
        self.assertEqual(0.07, results.solver.system_time)
        self.assertEqual(0.07, results.solver.wallclock_time)
        self.assertEqual(
            TerminationCondition.optimal, results.solver.termination_condition
        )
        self.assertEqual(
            'Model was solved to optimality (subject to tolerances), and an optimal solution is available.',
            results.solver.termination_message,
        )
        self.assertEqual(
            results.solver.statistics.branch_and_bound.number_of_bounded_subproblems, 0
        )
        self.assertEqual(
            results.solver.statistics.branch_and_bound.number_of_created_subproblems, 0
        )
        self.assertEqual(results.solver.statistics.black_box.number_of_iterations, 0)

    def test_max_evaluations(self):
        """
        solver_kwargs={'options': {'maxNodes': 1}}
        """
        lp_file = 'max_evaluations.out.lp'
        results = self.opt.solve(os.path.join(data_dir, lp_file))

        self.assertEqual(1.2052223, results.problem.lower_bound)
        self.assertEqual(1.20645976, results.problem.upper_bound)
        self.assertEqual(SolverStatus.aborted, results.solver.status)
        self.assertEqual(0.16, results.solver.system_time)
        self.assertEqual(0.18, results.solver.wallclock_time)
        self.assertEqual(
            TerminationCondition.maxEvaluations, results.solver.termination_condition
        )
        self.assertEqual(
            'Optimization terminated because the total number of branch-and-cut nodes explored exceeded the value '
            'specified in the maxNodes parameter',
            results.solver.termination_message,
        )
        self.assertEqual(
            results.solver.statistics.branch_and_bound.number_of_bounded_subproblems, 1
        )
        self.assertEqual(
            results.solver.statistics.branch_and_bound.number_of_created_subproblems, 1
        )
        self.assertEqual(results.solver.statistics.black_box.number_of_iterations, 602)

    def test_fix_parsing_bug(self):
        """
        The test wasn't generated using the method in the class docstring
        See https://github.com/Pyomo/pyomo/issues/1001
        """
        lp_file = 'fix_parsing_bug.out.lp'
        results = self.opt.solve(os.path.join(data_dir, lp_file))

        if self.opt.version() < (2, 10, 2):
            self.assertEqual(3.0, results.problem.lower_bound)
            self.assertEqual(3.0, results.problem.upper_bound)
        else:
            self.assertEqual(-3.0, results.problem.lower_bound)
            self.assertEqual(-3.0, results.problem.upper_bound)
        self.assertEqual(SolverStatus.aborted, results.solver.status)
        self.assertEqual(0.08, results.solver.system_time)
        self.assertEqual(0.09, results.solver.wallclock_time)
        self.assertEqual(
            TerminationCondition.other, results.solver.termination_condition
        )
        self.assertEqual(
            'Optimization terminated because the number of solutions found reached the value specified in the '
            'maxSolutions parameter.',
            results.solver.termination_message,
        )
        self.assertEqual(
            results.solver.statistics.branch_and_bound.number_of_bounded_subproblems, 0
        )
        self.assertEqual(
            results.solver.statistics.branch_and_bound.number_of_created_subproblems, 0
        )
        self.assertEqual(results.solver.statistics.black_box.number_of_iterations, 0)

    def test_process_logfile(self):
        cbc_shell = CBCSHELL()
        cbc_shell._log_file = os.path.join(data_dir, 'test5_timeout_cbc.txt')
        results = cbc_shell.process_logfile()
        self.assertEqual(results.solution.gap, 0.01)
        self.assertEqual(
            results.solver.statistics.black_box.number_of_iterations, 50364
        )
        self.assertEqual(results.solver.system_time, 2.01)
        self.assertEqual(
            results.solver.statistics.branch_and_bound.number_of_created_subproblems,
            34776,
        )

    def test_process_logfile_gap_inf(self):
        cbc_shell = CBCSHELL()
        cbc_shell._log_file = os.path.join(data_dir, 'test5_timeout_cbc_gap.txt')
        results = cbc_shell.process_logfile()
        self.assertEqual(results.solution.gap, float('inf'))
        self.assertEqual(
            results.solver.statistics.black_box.number_of_iterations, 50364
        )
        self.assertEqual(results.solver.system_time, 2.01)
        self.assertEqual(
            results.solver.statistics.branch_and_bound.number_of_created_subproblems,
            34776,
        )


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common import unittest
import pyomo.environ as pyo
from pyomo.contrib.solver.common.util import collect_vars_and_named_exprs, get_objective
from pyomo.opt.results.solver import (
    SolverStatus as LegacySolverStatus,
    TerminationCondition as LegacyTerminationCondition,
)
from pyomo.opt.results import SolverResults as LegacySolverResults
from pyomo.contrib.solver.common.results import Results, SolutionStatus
from typing import Callable
from pyomo.common.gsl import find_GSL


class TestGenericUtils(unittest.TestCase):
    def basics_helper(self, collector: Callable, *args):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.z = pyo.Var()
        m.E = pyo.Expression(expr=2 * m.z + 1)
        m.y.fix(3)
        e = m.x * m.y + m.x * m.E
        named_exprs, var_list, fixed_vars, external_funcs = collector(e, *args)
        self.assertEqual([m.E], named_exprs)
        self.assertEqual([m.x, m.y, m.z], var_list)
        self.assertEqual([m.y], fixed_vars)
        self.assertEqual([], external_funcs)

    def test_collect_vars_basics(self):
        self.basics_helper(collect_vars_and_named_exprs)

    def external_func_helper(self, collector: Callable, *args):
        DLL = find_GSL()
        if not DLL:
            self.skipTest('Could not find amplgsl.dll library')

        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.z = pyo.Var()
        m.hypot = pyo.ExternalFunction(library=DLL, function='gsl_hypot')
        func = m.hypot(m.x, m.x * m.y)
        m.E = pyo.Expression(expr=2 * func)
        m.y.fix(3)
        e = m.z + m.x * m.E
        named_exprs, var_list, fixed_vars, external_funcs = collector(e, *args)
        self.assertEqual([m.E], named_exprs)
        self.assertEqual([m.z, m.x, m.y], var_list)
        self.assertEqual([m.y], fixed_vars)
        self.assertEqual([func], external_funcs)

    def test_collect_vars_external(self):
        self.external_func_helper(collect_vars_and_named_exprs)

    def simple_model(self):
        model = pyo.ConcreteModel()
        model.x = pyo.Var([1, 2], domain=pyo.NonNegativeReals)
        model.OBJ = pyo.Objective(expr=2 * model.x[1] + 3 * model.x[2])
        model.Constraint1 = pyo.Constraint(expr=3 * model.x[1] + 4 * model.x[2] >= 1)
        return model

    def test_get_objective_success(self):
        model = self.simple_model()
        self.assertEqual(model.OBJ, get_objective(model))

    def test_get_objective_raise(self):
        model = self.simple_model()
        model.OBJ2 = pyo.Objective(expr=model.x[1] - 4 * model.x[2])
        with self.assertRaises(ValueError):
            get_objective(model)

    def test_check_optimal_termination_new_interface(self):
        results = Results()
        results.solution_status = SolutionStatus.optimal
        # Optimal solution
        self.assertTrue(pyo.check_optimal_termination(results))
        # Anything other than optimal
        results.solution_status = SolutionStatus.noSolution
        self.assertFalse(pyo.check_optimal_termination(results))

    def test_check_optimal_termination_condition_legacy_interface(self):
        results = LegacySolverResults()
        results.solver.status = LegacySolverStatus.ok
        results.solver.termination_condition = LegacyTerminationCondition.optimal
        # Both items satisfied
        self.assertTrue(pyo.check_optimal_termination(results))
        # Termination condition not satisfied
        results.solver.termination_condition = LegacyTerminationCondition.unknown
        self.assertFalse(pyo.check_optimal_termination(results))
        # Both not satisfied
        results.solver.termination_condition = LegacySolverStatus.aborted
        self.assertFalse(pyo.check_optimal_termination(results))

    def test_assert_optimal_termination_new_interface(self):
        results = Results()
        results.solution_status = SolutionStatus.optimal
        pyo.assert_optimal_termination(results)
        # No solution
        results.solution_status = SolutionStatus.noSolution
        with self.assertRaises(RuntimeError):
            pyo.assert_optimal_termination(results)

    def test_assert_optimal_termination_legacy_interface(self):
        results = LegacySolverResults()
        results.solver.status = LegacySolverStatus.ok
        results.solver.termination_condition = LegacyTerminationCondition.optimal
        pyo.assert_optimal_termination(results)
        # Termination condition not satisfied
        results.solver.termination_condition = LegacyTerminationCondition.unknown
        with self.assertRaises(RuntimeError):
            pyo.assert_optimal_termination(results)
        # Both not satisfied
        results.solver.termination_condition = LegacySolverStatus.aborted
        with self.assertRaises(RuntimeError):
            pyo.assert_optimal_termination(results)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# This module was originally developed as part of the PyUtilib project
# Copyright (c) 2008 Sandia Corporation.
# This software is distributed under the BSD License.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
# ____________________________________________________________________________________

import gc
from pyomo.common.multithread import MultiThreadWrapper


class __PauseGCCompanion:
    def __init__(self):
        self._stack_depth = 0


PauseGCCompanion: __PauseGCCompanion = MultiThreadWrapper(__PauseGCCompanion)


# PauseGC is a class for clean, scoped management of the Python
# garbage collector.  To disable the GC for the duration of a
# scoped block use PauseGC in combination with the Python 'with'
# statement. E.g.,
#
# def my_func():
#    with PauseGC() as pgc:
#       [...]
#
# When the instance falls out of scope (by termination or exception),
# the GC will be re-enabled (if it was not initially disabled).  It is
# safe to nest instances of PauseGC That is, you don't have to worry
# if an outer function/method has its own instance of PauseGC.
class PauseGC:
    __slots__ = ("reenable_gc", "stack_pointer")

    def __init__(self):
        self.stack_pointer = None
        self.reenable_gc = None

    def __enter__(self):
        if self.stack_pointer:
            raise RuntimeError(
                "Entering PauseGC context manager that was already entered."
            )
        PauseGCCompanion._stack_depth += 1
        self.stack_pointer = PauseGCCompanion._stack_depth
        self.reenable_gc = gc.isenabled()
        if self.reenable_gc:
            gc.disable()
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def close(self):
        if not self.stack_pointer:
            return
        if PauseGCCompanion._stack_depth:
            if PauseGCCompanion._stack_depth != self.stack_pointer:
                raise RuntimeError(
                    "Exiting PauseGC context manager out of order: there "
                    "are other active PauseGC context managers that were "
                    "entered after this context manager and have not yet "
                    "been exited."
                )
            PauseGCCompanion._stack_depth -= 1
            self.stack_pointer = None
        if self.reenable_gc:
            gc.enable()
        self.reenable_gc = None



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
import re
import sys

from pyomo.common.tempfiles import TempfileManager
from pyomo.common.collections import ComponentSet, ComponentMap, Bunch
from pyomo.core.base import Suffix, Var, Constraint, SOSConstraint, Objective
from pyomo.core.expr.numvalue import is_fixed
from pyomo.core.expr.numvalue import value
from pyomo.core.staleflag import StaleFlagManager
from pyomo.repn import generate_standard_repn
from pyomo.solvers.plugins.solvers.direct_solver import DirectSolver
from pyomo.solvers.plugins.solvers.direct_or_persistent_solver import (
    DirectOrPersistentSolver,
)
from pyomo.core.kernel.objective import minimize, maximize
from pyomo.opt.results.results_ import SolverResults
from pyomo.opt.results.solution import Solution, SolutionStatus
from pyomo.opt.results.solver import TerminationCondition, SolverStatus
from pyomo.opt.base import SolverFactory
import time

logger = logging.getLogger('pyomo.solvers')


class DegreeError(ValueError):
    pass


class _CplexExpr:
    def __init__(
        self,
        variables,
        coefficients,
        offset=None,
        q_variables1=None,
        q_variables2=None,
        q_coefficients=None,
    ):
        self.variables = variables
        self.coefficients = coefficients
        self.offset = offset or 0.0
        self.q_variables1 = q_variables1 or []
        self.q_variables2 = q_variables2 or []
        self.q_coefficients = [float(coef) for coef in q_coefficients or []]


def _is_numeric(x):
    try:
        float(x)
    except ValueError:
        return False
    return True


class _VariableData:
    def __init__(self, solver_model):
        self._solver_model = solver_model
        self.lb = []
        self.ub = []
        self.types = []
        self.names = []

    def add(self, lb, ub, type_, name):
        self.lb.append(lb)
        self.ub.append(ub)
        self.types.append(type_)
        self.names.append(name)

    def store_in_cplex(self):
        self._solver_model.variables.add(
            lb=self.lb, ub=self.ub, types=self.types, names=self.names
        )


class _LinearConstraintData:
    def __init__(self, solver_model):
        self._solver_model = solver_model
        self.lin_expr = []
        self.senses = []
        self.rhs = []
        self.range_values = []
        self.names = []

    def add(self, cplex_expr, sense, rhs, range_values, name):
        self.lin_expr.append([cplex_expr.variables, cplex_expr.coefficients])
        self.senses.append(sense)
        self.rhs.append(rhs)
        self.range_values.append(range_values)
        self.names.append(name)

    def store_in_cplex(self):
        self._solver_model.linear_constraints.add(
            lin_expr=self.lin_expr,
            senses=self.senses,
            rhs=self.rhs,
            range_values=self.range_values,
            names=self.names,
        )


@SolverFactory.register('cplex_direct', doc='Direct python interface to CPLEX')
class CPLEXDirect(DirectSolver):
    def __init__(self, **kwds):
        kwds['type'] = 'cplexdirect'
        DirectSolver.__init__(self, **kwds)
        self._init()
        self._wallclock_time = None
        self._pyomo_var_to_ndx_map = ComponentMap()
        self._ndx_count = 0

    def _init(self):
        try:
            import cplex

            self._cplex = cplex
            self._python_api_exists = True
            self._version = tuple(
                int(k) for k in self._cplex.Cplex().get_version().split('.')
            )
            while len(self._version) < 4:
                self._version += (0,)
            self._version = tuple(int(i) for i in self._version[:4])
            self._version_major = self._version[0]
        except ImportError:
            self._python_api_exists = False
        except Exception as e:
            # other forms of exceptions can be thrown by the gurobi python
            # import. for example, a gurobipy.GurobiError exception is thrown
            # if all tokens for Gurobi are already in use. assuming, of
            # course, the license is a token license. unfortunately, you can't
            # import without a license, which means we can't test for the
            # exception above!
            print("Import of cplex failed - cplex message=" + str(e) + "\n")
            self._python_api_exists = False

        self._range_constraints = set()

        self._max_constraint_degree = 2
        self._max_obj_degree = 2

        # Note: Undefined capabilities default to None
        self._capabilities.linear = True
        self._capabilities.quadratic_objective = True
        self._capabilities.quadratic_constraint = True
        self._capabilities.integer = True
        self._capabilities.sos1 = True
        self._capabilities.sos2 = True

    def _apply_solver(self):
        StaleFlagManager.mark_all_as_stale()

        # In recent versions of CPLEX it is helpful to manually open the
        # log file and then explicitly close it after CPLEX is finished.
        # This ensures that the file is closed (and unlocked) on Windows
        # before the TempfileManager (or user) attempts to delete the
        # log file.  Passing in an opened file object is supported at
        # least as far back as CPLEX 12.5.1 [the oldest version
        # supported by IBM as of 1 Oct 2020]
        if self.version() >= (12, 5, 1) and isinstance(self._log_file, str):
            _log_file = (open(self._log_file, 'a'),)
            _close_log_file = True
        else:
            _log_file = (self._log_file,)
            _close_log_file = False
        if self._tee:

            def _process_stream(arg):
                sys.stdout.write(arg)
                return arg

            _log_file += (_process_stream,)
        try:
            self._solver_model.set_results_stream(*_log_file)
            if self._keepfiles:
                print("Solver log file: " + self._log_file)

            obj_degree = self._objective.polynomial_degree()
            if obj_degree is None or obj_degree > 2:
                raise DegreeError(
                    'CPLEXDirect does not support expressions of degree {0}.'.format(
                        obj_degree
                    )
                )
            elif obj_degree == 2:
                quadratic_objective = True
            else:
                quadratic_objective = False

            num_integer_vars = self._solver_model.variables.get_num_integer()
            num_binary_vars = self._solver_model.variables.get_num_binary()
            num_sos = self._solver_model.SOS.get_num()

            if self._solver_model.quadratic_constraints.get_num() != 0:
                quadratic_cons = True
            else:
                quadratic_cons = False

            if (num_integer_vars + num_binary_vars + num_sos) > 0:
                integer = True
            else:
                integer = False

            if integer:
                if quadratic_cons:
                    self._solver_model.set_problem_type(
                        self._solver_model.problem_type.MIQCP
                    )
                elif quadratic_objective:
                    self._solver_model.set_problem_type(
                        self._solver_model.problem_type.MIQP
                    )
                else:
                    self._solver_model.set_problem_type(
                        self._solver_model.problem_type.MILP
                    )
            else:
                if quadratic_cons:
                    self._solver_model.set_problem_type(
                        self._solver_model.problem_type.QCP
                    )
                elif quadratic_objective:
                    self._solver_model.set_problem_type(
                        self._solver_model.problem_type.QP
                    )
                else:
                    self._solver_model.set_problem_type(
                        self._solver_model.problem_type.LP
                    )

            # if the user specifies a 'mipgap'
            # set cplex's mip.tolerances.mipgap
            if self.options.mipgap is not None:
                self._solver_model.parameters.mip.tolerances.mipgap.set(
                    float(self.options.mipgap)
                )

            for key, option in self.options.items():
                if key == 'mipgap':  # handled above
                    continue
                opt_cmd = self._solver_model.parameters
                key_pieces = key.split('_')
                for key_piece in key_pieces:
                    opt_cmd = getattr(opt_cmd, key_piece)
                # When options come from the pyomo command, all
                # values are string types, so we try to cast
                # them to a numeric value in the event that
                # setting the parameter fails.
                try:
                    opt_cmd.set(option)
                except self._cplex.exceptions.CplexError:
                    # we place the exception handling for
                    # checking the cast of option to a float in
                    # another function so that we can simply
                    # call raise here instead of except
                    # TypeError as e / raise e, because the
                    # latter does not preserve the Cplex stack
                    # trace
                    if not _is_numeric(option):
                        raise
                    opt_cmd.set(float(option))

            t0 = time.time()
            self._solver_model.solve()
            t1 = time.time()
            self._wallclock_time = t1 - t0
        finally:
            self._solver_model.set_results_stream(None)
            if _close_log_file:
                _log_file[0].close()

        # FIXME: can we get a return code indicating if CPLEX had a significant failure?
        return Bunch(rc=None, log=None)

    def _get_expr_from_pyomo_repn(self, repn, max_degree=2):
        degree = repn.polynomial_degree()
        if degree is None or degree > max_degree:
            raise DegreeError(
                "CPLEXDirect does not support expressions of degree {0}.".format(degree)
            )

        referenced_vars = ComponentSet(repn.linear_vars)

        q_coefficients = []
        q_variables1 = []
        q_variables2 = []
        for i, v in enumerate(repn.quadratic_vars):
            x, y = v
            q_coefficients.append(repn.quadratic_coefs[i])
            q_variables1.append(self._pyomo_var_to_ndx_map[x])
            q_variables2.append(self._pyomo_var_to_ndx_map[y])
            referenced_vars.add(x)
            referenced_vars.add(y)

        return (
            _CplexExpr(
                variables=[self._pyomo_var_to_ndx_map[var] for var in repn.linear_vars],
                coefficients=repn.linear_coefs,
                offset=repn.constant,
                q_variables1=q_variables1,
                q_variables2=q_variables2,
                q_coefficients=q_coefficients,
            ),
            referenced_vars,
        )

    def _get_expr_from_pyomo_expr(self, expr, max_degree=2):
        if max_degree == 2:
            repn = generate_standard_repn(expr, quadratic=True)
        else:
            repn = generate_standard_repn(expr, quadratic=False)

        try:
            cplex_expr, referenced_vars = self._get_expr_from_pyomo_repn(
                repn, max_degree
            )
        except DegreeError as e:
            msg = e.args[0]
            msg += '\nexpr: {0}'.format(expr)
            raise DegreeError(msg)

        return cplex_expr, referenced_vars

    def _cplex_lb_ub_from_var(self, var):
        if var.is_fixed():
            val = var.value
            return val, val
        if var.has_lb():
            lb = value(var.lb)
        else:
            lb = -self._cplex.infinity
        if var.has_ub():
            ub = value(var.ub)
        else:
            ub = self._cplex.infinity
        return lb, ub

    def _add_var(self, var, var_data=None):
        varname = self._symbol_map.getSymbol(var, self._labeler)
        vtype = self._cplex_vtype_from_var(var)
        lb, ub = self._cplex_lb_ub_from_var(var)

        cplex_var_data = (
            _VariableData(self._solver_model) if var_data is None else var_data
        )
        cplex_var_data.add(lb=lb, ub=ub, type_=vtype, name=varname)
        if var_data is None:
            cplex_var_data.store_in_cplex()

        self._pyomo_var_to_solver_var_map[var] = varname
        self._solver_var_to_pyomo_var_map[varname] = var
        self._pyomo_var_to_ndx_map[var] = self._ndx_count
        self._ndx_count += 1
        self._referenced_variables[var] = 0

    def _set_instance(self, model, kwds={}):
        self._pyomo_var_to_ndx_map = ComponentMap()
        self._ndx_count = 0
        self._range_constraints = set()
        DirectOrPersistentSolver._set_instance(self, model, kwds)
        try:
            self._solver_model = self._cplex.Cplex()
        except Exception:
            e = sys.exc_info()[1]
            msg = (
                "Unable to create CPLEX model. "
                "Have you installed the Python "
                "bindings for CPLEX?\n\n\t" + "Error message: {0}".format(e)
            )
            raise Exception(msg)

        self._add_block(model)

        for var, n_ref in self._referenced_variables.items():
            if n_ref != 0:
                if var.fixed:
                    if not self._output_fixed_variable_bounds:
                        raise ValueError(
                            "Encountered a fixed variable (%s) inside "
                            "an active objective or constraint "
                            "expression on model %s, which is usually "
                            "indicative of a preprocessing error. Use "
                            "the IO-option 'output_fixed_variable_bounds=True' "
                            "to suppress this error and fix the variable "
                            "by overwriting its bounds in the CPLEX instance."
                            % (var.name, self._pyomo_model.name)
                        )

    def _add_block(self, block):
        var_data = _VariableData(self._solver_model)
        for var in block.component_data_objects(
            ctype=Var, descend_into=True, active=True, sort=True
        ):
            self._add_var(var, var_data)
        var_data.store_in_cplex()

        lin_con_data = _LinearConstraintData(self._solver_model)
        for sub_block in block.block_data_objects(descend_into=True, active=True):
            for con in sub_block.component_data_objects(
                ctype=Constraint, descend_into=False, active=True, sort=True
            ):
                if not con.has_lb() and not con.has_ub():
                    assert not con.equality
                    continue  # non-binding, so skip

                self._add_constraint(con, lin_con_data)

            for con in sub_block.component_data_objects(
                ctype=SOSConstraint, descend_into=False, active=True, sort=True
            ):
                self._add_sos_constraint(con)

            obj_counter = 0
            for obj in sub_block.component_data_objects(
                ctype=Objective, descend_into=False, active=True
            ):
                obj_counter += 1
                if obj_counter > 1:
                    raise ValueError(
                        "Solver interface does not support multiple objectives."
                    )
                self._set_objective(obj)
        lin_con_data.store_in_cplex()

    def _add_constraint(self, con, lin_con_data=None):
        if not con.active:
            return None

        if self._skip_trivial_constraints and is_fixed(con.body):
            return None

        conname = self._symbol_map.getSymbol(con, self._labeler)

        if con._linear_canonical_form:
            cplex_expr, referenced_vars = self._get_expr_from_pyomo_repn(
                con.canonical_form(), self._max_constraint_degree
            )
        else:
            cplex_expr, referenced_vars = self._get_expr_from_pyomo_expr(
                con.body, self._max_constraint_degree
            )

        if con.has_lb() and not is_fixed(con.lower):
            raise ValueError(
                "Lower bound of constraint {0} is not constant.".format(con)
            )

        if con.has_ub() and not is_fixed(con.upper):
            raise ValueError(
                "Upper bound of constraint {0} is not constant.".format(con)
            )

        range_ = 0.0
        if con.equality:
            sense = "E"
            rhs = value(con.lower) - cplex_expr.offset
        elif con.has_lb() and con.has_ub():
            sense = "R"
            lb = value(con.lower)
            ub = value(con.upper)
            rhs = ub - cplex_expr.offset
            range_ = lb - ub
            self._range_constraints.add(con)
        elif con.has_lb():
            sense = "G"
            rhs = value(con.lower) - cplex_expr.offset
        elif con.has_ub():
            sense = "L"
            rhs = value(con.upper) - cplex_expr.offset
        else:
            raise ValueError(
                "Constraint does not have a lower "
                "or an upper bound: {0} \n".format(con)
            )

        if len(cplex_expr.q_coefficients) == 0:
            cplex_lin_con_data = (
                _LinearConstraintData(self._solver_model)
                if lin_con_data is None
                else lin_con_data
            )
            cplex_lin_con_data.add(cplex_expr, sense, rhs, range_, conname)
            if lin_con_data is None:
                cplex_lin_con_data.store_in_cplex()
        else:
            if sense == 'R':
                raise ValueError(
                    "The CPLEXDirect interface does not "
                    "support quadratic range constraints: "
                    "{0}".format(con)
                )
            self._solver_model.quadratic_constraints.add(
                lin_expr=[cplex_expr.variables, cplex_expr.coefficients],
                quad_expr=[
                    cplex_expr.q_variables1,
                    cplex_expr.q_variables2,
                    cplex_expr.q_coefficients,
                ],
                sense=sense,
                rhs=rhs,
                name=conname,
            )

        for var in referenced_vars:
            self._referenced_variables[var] += 1
        self._vars_referenced_by_con[con] = referenced_vars
        self._pyomo_con_to_solver_con_map[con] = conname
        self._solver_con_to_pyomo_con_map[conname] = con

    def _add_sos_constraint(self, con):
        if not con.active:
            return None

        conname = self._symbol_map.getSymbol(con, self._labeler)
        level = con.level
        if level == 1:
            sos_type = self._solver_model.SOS.type.SOS1
        elif level == 2:
            sos_type = self._solver_model.SOS.type.SOS2
        else:
            raise ValueError(
                "Solver does not support SOS level {0} constraints".format(level)
            )

        cplex_vars = []
        weights = []

        self._vars_referenced_by_con[con] = ComponentSet()

        if hasattr(con, 'get_items'):
            # aml sos constraint
            sos_items = list(con.get_items())
        else:
            # kernel sos constraint
            sos_items = list(con.items())

        for v, w in sos_items:
            self._vars_referenced_by_con[con].add(v)
            cplex_vars.append(self._pyomo_var_to_solver_var_map[v])
            self._referenced_variables[v] += 1
            weights.append(w)

        self._solver_model.SOS.add(
            type=sos_type, SOS=[cplex_vars, weights], name=conname
        )
        self._pyomo_con_to_solver_con_map[con] = conname
        self._solver_con_to_pyomo_con_map[conname] = con

    def _cplex_vtype_from_var(self, var):
        """
        This function takes a pyomo variable and returns the appropriate gurobi variable type
        :param var: pyomo.core.base.var.Var
        :return: gurobipy.GRB.CONTINUOUS or gurobipy.GRB.BINARY or gurobipy.GRB.INTEGER
        """
        if var.is_binary():
            vtype = self._solver_model.variables.type.binary
        elif var.is_integer():
            vtype = self._solver_model.variables.type.integer
        elif var.is_continuous():
            vtype = self._solver_model.variables.type.continuous
        else:
            raise ValueError(
                'Variable domain type is not recognized for {0}'.format(var.domain)
            )
        return vtype

    def _set_objective(self, obj):
        if self._objective is not None:
            for var in self._vars_referenced_by_obj:
                self._referenced_variables[var] -= 1
            self._vars_referenced_by_obj = ComponentSet()
            self._objective = None

        if obj.active is False:
            raise ValueError('Cannot add inactive objective to solver.')

        if obj.sense == minimize:
            sense = self._solver_model.objective.sense.minimize
        elif obj.sense == maximize:
            sense = self._solver_model.objective.sense.maximize
        else:
            raise ValueError('Objective sense is not recognized: {0}'.format(obj.sense))

        cplex_expr, referenced_vars = self._get_expr_from_pyomo_expr(
            obj.expr, self._max_obj_degree
        )
        # CPLEX actually uses x'Qx/2 in the objective, as the
        # off-diagonal entries appear in both the lower triangle and the
        # upper triangle (i.e., c*x1*x2 and c*x2*x1).  However, since
        # the diagonal entries only appear once, we need to double them.
        for i, v1 in enumerate(cplex_expr.q_variables1):
            if v1 == cplex_expr.q_variables2[i]:
                cplex_expr.q_coefficients[i] *= 2

        for var in referenced_vars:
            self._referenced_variables[var] += 1

        self._solver_model.objective.set_sense(sense)
        if hasattr(self._solver_model.objective, 'set_offset'):
            self._solver_model.objective.set_offset(cplex_expr.offset)

        linear_objective_already_exists = any(self._solver_model.objective.get_linear())
        quadratic_objective_already_exists = (
            self._solver_model.objective.get_num_quadratic_nonzeros()
        )

        contains_linear_terms = any(cplex_expr.coefficients)
        contains_quadratic_terms = any(cplex_expr.q_coefficients)
        num_cols = len(self._pyomo_var_to_solver_var_map)

        if linear_objective_already_exists or contains_linear_terms:
            self._solver_model.objective.set_linear([(i, 0.0) for i in range(num_cols)])

            if contains_linear_terms:
                self._solver_model.objective.set_linear(
                    list(zip(cplex_expr.variables, cplex_expr.coefficients))
                )

        if quadratic_objective_already_exists or contains_quadratic_terms:
            self._solver_model.objective.set_quadratic([0.0] * num_cols)

            if contains_quadratic_terms:
                self._solver_model.objective.set_quadratic_coefficients(
                    list(
                        zip(
                            cplex_expr.q_variables1,
                            cplex_expr.q_variables2,
                            cplex_expr.q_coefficients,
                        )
                    )
                )
        self._objective = obj
        self._vars_referenced_by_obj = referenced_vars

    def _postsolve(self):
        # the only suffixes that we extract from CPLEX are
        # constraint duals, constraint slacks, and variable
        # reduced-costs. scan through the solver suffix list
        # and throw an exception if the user has specified
        # any others.
        extract_duals = False
        extract_slacks = False
        extract_reduced_costs = False
        for suffix in self._suffixes:
            flag = False
            if re.match(suffix, "dual"):
                extract_duals = True
                flag = True
            if re.match(suffix, "slack"):
                extract_slacks = True
                flag = True
            if re.match(suffix, "rc"):
                extract_reduced_costs = True
                flag = True
            if not flag:
                raise RuntimeError(
                    "***The cplex_direct solver plugin cannot extract solution suffix="
                    + suffix
                )

        cpxprob = self._solver_model
        status = cpxprob.solution.get_status()

        if cpxprob.get_problem_type() in [
            cpxprob.problem_type.MILP,
            cpxprob.problem_type.MIQP,
            cpxprob.problem_type.MIQCP,
        ]:
            if extract_reduced_costs:
                logger.warning("Cannot get reduced costs for MIP.")
            if extract_duals:
                logger.warning("Cannot get duals for MIP.")
            extract_reduced_costs = False
            extract_duals = False

        self.results = SolverResults()
        soln = Solution()

        self.results.solver.name = "CPLEX {0}".format(cpxprob.get_version())
        self.results.solver.wallclock_time = self._wallclock_time

        if status in [1, 101, 102]:
            self.results.solver.status = SolverStatus.ok
            self.results.solver.termination_condition = TerminationCondition.optimal
            soln.status = SolutionStatus.optimal
        elif status in [2, 40, 118, 133, 134]:
            self.results.solver.status = SolverStatus.warning
            self.results.solver.termination_condition = TerminationCondition.unbounded
            soln.status = SolutionStatus.unbounded
        elif status in [4, 119, 134]:
            # Note: status of 4 means infeasible or unbounded
            #       and 119 means MIP infeasible or unbounded
            self.results.solver.status = SolverStatus.warning
            self.results.solver.termination_condition = (
                TerminationCondition.infeasibleOrUnbounded
            )
            soln.status = SolutionStatus.unsure
        elif status in [3, 103]:
            self.results.solver.status = SolverStatus.warning
            self.results.solver.termination_condition = TerminationCondition.infeasible
            soln.status = SolutionStatus.infeasible
        elif status in [10]:
            self.results.solver.status = SolverStatus.aborted
            self.results.solver.termination_condition = (
                TerminationCondition.maxIterations
            )
            soln.status = SolutionStatus.stoppedByLimit
        elif status in [11, 25, 107, 131]:
            self.results.solver.status = SolverStatus.aborted
            self.results.solver.termination_condition = (
                TerminationCondition.maxTimeLimit
            )
            soln.status = SolutionStatus.stoppedByLimit
        else:
            self.results.solver.status = SolverStatus.error
            self.results.solver.termination_condition = TerminationCondition.error
            soln.status = SolutionStatus.error

        if cpxprob.objective.get_sense() == cpxprob.objective.sense.minimize:
            self.results.problem.sense = minimize
        elif cpxprob.objective.get_sense() == cpxprob.objective.sense.maximize:
            self.results.problem.sense = maximize
        else:
            raise RuntimeError(
                'Unrecognized cplex objective sense: {0}'.format(
                    cpxprob.objective.get_sense()
                )
            )

        self.results.problem.upper_bound = None
        self.results.problem.lower_bound = None
        if cpxprob.solution.get_solution_type() != cpxprob.solution.type.none:
            if (
                cpxprob.variables.get_num_binary() + cpxprob.variables.get_num_integer()
            ) == 0:
                self.results.problem.upper_bound = (
                    cpxprob.solution.get_objective_value()
                )
                self.results.problem.lower_bound = (
                    cpxprob.solution.get_objective_value()
                )
            elif cpxprob.objective.get_sense() == cpxprob.objective.sense.minimize:
                self.results.problem.upper_bound = (
                    cpxprob.solution.get_objective_value()
                )
                self.results.problem.lower_bound = (
                    cpxprob.solution.MIP.get_best_objective()
                )
            else:
                assert cpxprob.objective.get_sense() == cpxprob.objective.sense.maximize
                self.results.problem.upper_bound = (
                    cpxprob.solution.MIP.get_best_objective()
                )
                self.results.problem.lower_bound = (
                    cpxprob.solution.get_objective_value()
                )

        try:
            soln.gap = (
                self.results.problem.upper_bound - self.results.problem.lower_bound
            )
        except TypeError:
            soln.gap = None

        self.results.problem.name = cpxprob.get_problem_name()
        assert cpxprob.indicator_constraints.get_num() == 0
        self.results.problem.number_of_constraints = (
            cpxprob.linear_constraints.get_num()
            + cpxprob.quadratic_constraints.get_num()
            + cpxprob.SOS.get_num()
        )
        self.results.problem.number_of_nonzeros = None
        self.results.problem.number_of_variables = cpxprob.variables.get_num()
        self.results.problem.number_of_binary_variables = (
            cpxprob.variables.get_num_binary()
        )
        self.results.problem.number_of_integer_variables = (
            cpxprob.variables.get_num_integer()
        )
        assert cpxprob.variables.get_num_semiinteger() == 0
        assert cpxprob.variables.get_num_semicontinuous() == 0
        self.results.problem.number_of_continuous_variables = (
            cpxprob.variables.get_num()
            - cpxprob.variables.get_num_binary()
            - cpxprob.variables.get_num_integer()
        )
        self.results.problem.number_of_objectives = 1

        # only try to get objective and variable values if a solution exists
        if self._save_results:
            """
            This code in this if statement is only needed for backwards compatibility. It is more efficient to set
            _save_results to False and use load_vars, load_duals, etc.
            """
            if cpxprob.solution.get_solution_type() > 0:
                soln_variables = soln.variable
                soln_constraints = soln.constraint

                var_names = self._solver_model.variables.get_names()
                assert set(var_names) == set(self._pyomo_var_to_solver_var_map.values())
                var_vals = self._solver_model.solution.get_values()
                for name, val in zip(var_names, var_vals):
                    pyomo_var = self._solver_var_to_pyomo_var_map[name]
                    if self._referenced_variables[pyomo_var] > 0:
                        soln_variables[name] = {"Value": val}

                if extract_reduced_costs:
                    reduced_costs = self._solver_model.solution.get_reduced_costs(
                        var_names
                    )
                    for i, name in enumerate(var_names):
                        pyomo_var = self._solver_var_to_pyomo_var_map[name]
                        if self._referenced_variables[pyomo_var] > 0:
                            soln_variables[name]["Rc"] = reduced_costs[i]

                if extract_slacks:
                    for con_name in self._solver_model.linear_constraints.get_names():
                        soln_constraints[con_name] = {}
                    for (
                        con_name
                    ) in self._solver_model.quadratic_constraints.get_names():
                        soln_constraints[con_name] = {}
                elif extract_duals:
                    # CPLEX PYTHON API DOES NOT SUPPORT QUADRATIC DUAL COLLECTION
                    for con_name in self._solver_model.linear_constraints.get_names():
                        soln_constraints[con_name] = {}

                if extract_duals:
                    dual_values = self._solver_model.solution.get_dual_values()
                    for i, con_name in enumerate(
                        self._solver_model.linear_constraints.get_names()
                    ):
                        soln_constraints[con_name]["Dual"] = dual_values[i]

                if extract_slacks:
                    linear_slacks = self._solver_model.solution.get_linear_slacks()
                    quadratic_slacks = (
                        self._solver_model.solution.get_quadratic_slacks()
                    )
                    for i, con_name in enumerate(
                        self._solver_model.linear_constraints.get_names()
                    ):
                        pyomo_con = self._solver_con_to_pyomo_con_map[con_name]
                        if pyomo_con in self._range_constraints:
                            R_ = self._solver_model.linear_constraints.get_range_values(
                                con_name
                            )
                            if R_ == 0:
                                soln_constraints[con_name]["Slack"] = linear_slacks[i]
                            else:
                                Ls_ = linear_slacks[i]
                                Us_ = R_ - Ls_
                                if abs(Us_) > abs(Ls_):
                                    soln_constraints[con_name]["Slack"] = Us_
                                else:
                                    soln_constraints[con_name]["Slack"] = -Ls_
                        else:
                            soln_constraints[con_name]["Slack"] = linear_slacks[i]
                    for i, con_name in enumerate(
                        self._solver_model.quadratic_constraints.get_names()
                    ):
                        soln_constraints[con_name]["Slack"] = quadratic_slacks[i]
        elif self._load_solutions:
            if cpxprob.solution.get_solution_type() > 0:
                self.load_vars()

                if extract_reduced_costs:
                    self._load_rc()

                if extract_duals:
                    self._load_duals()

                if extract_slacks:
                    self._load_slacks()

        self.results.solution.insert(soln)

        # finally, clean any temporary files registered with the temp file
        # manager, created populated *directly* by this plugin.
        TempfileManager.pop(remove=not self._keepfiles)

        return DirectOrPersistentSolver._postsolve(self)

    def warm_start_capable(self):
        return True

    def _warm_start(self):
        # here warm start means MIP start, which we can not add
        # if the problem type is not discrete
        cpxprob = self._solver_model
        if cpxprob.get_problem_type() in [
            cpxprob.problem_type.MILP,
            cpxprob.problem_type.MIQP,
            cpxprob.problem_type.MIQCP,
        ]:
            var_names = []
            var_values = []
            for pyomo_var, cplex_var in self._pyomo_var_to_solver_var_map.items():
                if pyomo_var.value is not None:
                    var_names.append(cplex_var)
                    var_values.append(value(pyomo_var))

            if len(var_names):
                self._solver_model.MIP_starts.add(
                    [var_names, var_values],
                    self._solver_model.MIP_starts.effort_level.auto,
                )

    def _load_vars(self, vars_to_load=None):
        var_map = self._pyomo_var_to_ndx_map
        if vars_to_load is None:
            vals = self._solver_model.solution.get_values()
            vars_to_load = var_map.keys()
        else:
            cplex_vars_to_load = [var_map[pyomo_var] for pyomo_var in vars_to_load]
            vals = self._solver_model.solution.get_values(cplex_vars_to_load)

        for pyomo_var, val in zip(vars_to_load, vals):
            if self._referenced_variables[pyomo_var] > 0:
                pyomo_var.set_value(val, skip_validation=True)

    def _load_rc(self, vars_to_load=None):
        if not hasattr(self._pyomo_model, 'rc'):
            self._pyomo_model.rc = Suffix(direction=Suffix.IMPORT)
        var_map = self._pyomo_var_to_solver_var_map
        ref_vars = self._referenced_variables
        rc = self._pyomo_model.rc
        if vars_to_load is None:
            vars_to_load = var_map.keys()

        cplex_vars_to_load = [var_map[pyomo_var] for pyomo_var in vars_to_load]
        vals = self._solver_model.solution.get_reduced_costs(cplex_vars_to_load)

        for i, pyomo_var in enumerate(vars_to_load):
            if ref_vars[pyomo_var] > 0:
                rc[pyomo_var] = vals[i]

    def _load_duals(self, cons_to_load=None):
        if not hasattr(self._pyomo_model, 'dual'):
            self._pyomo_model.dual = Suffix(direction=Suffix.IMPORT)
        con_map = self._pyomo_con_to_solver_con_map
        reverse_con_map = self._solver_con_to_pyomo_con_map
        dual = self._pyomo_model.dual

        if cons_to_load is None:
            linear_cons_to_load = self._solver_model.linear_constraints.get_names()
            vals = self._solver_model.solution.get_dual_values()
        else:
            cplex_cons_to_load = set([con_map[pyomo_con] for pyomo_con in cons_to_load])
            linear_cons_to_load = cplex_cons_to_load.intersection(
                set(self._solver_model.linear_constraints.get_names())
            )
            vals = self._solver_model.solution.get_dual_values(linear_cons_to_load)

        for i, cplex_con in enumerate(linear_cons_to_load):
            pyomo_con = reverse_con_map[cplex_con]
            dual[pyomo_con] = vals[i]

    def _load_slacks(self, cons_to_load=None):
        if not hasattr(self._pyomo_model, 'slack'):
            self._pyomo_model.slack = Suffix(direction=Suffix.IMPORT)
        con_map = self._pyomo_con_to_solver_con_map
        reverse_con_map = self._solver_con_to_pyomo_con_map
        slack = self._pyomo_model.slack

        if cons_to_load is None:
            linear_cons_to_load = self._solver_model.linear_constraints.get_names()
            linear_vals = self._solver_model.solution.get_linear_slacks()
            quadratic_cons_to_load = (
                self._solver_model.quadratic_constraints.get_names()
            )
            quadratic_vals = self._solver_model.solution.get_quadratic_slacks()
        else:
            cplex_cons_to_load = set([con_map[pyomo_con] for pyomo_con in cons_to_load])
            linear_cons_to_load = cplex_cons_to_load.intersection(
                set(self._solver_model.linear_constraints.get_names())
            )
            linear_vals = self._solver_model.solution.get_linear_slacks(
                linear_cons_to_load
            )
            quadratic_cons_to_load = cplex_cons_to_load.intersection(
                set(self._solver_model.quadratic_constraints.get_names())
            )
            quadratic_vals = self._solver_model.solution.get_quadratic_slacks(
                quadratic_cons_to_load
            )

        for i, cplex_con in enumerate(linear_cons_to_load):
            pyomo_con = reverse_con_map[cplex_con]
            if pyomo_con in self._range_constraints:
                R_ = self._solver_model.linear_constraints.get_range_values(cplex_con)
                if R_ == 0:
                    slack[pyomo_con] = linear_vals[i]
                else:
                    Ls_ = linear_vals[i]
                    Us_ = R_ - Ls_
                    if abs(Us_) > abs(Ls_):
                        slack[pyomo_con] = Us_
                    else:
                        slack[pyomo_con] = -Ls_
            else:
                slack[pyomo_con] = linear_vals[i]

        for i, cplex_con in enumerate(quadratic_cons_to_load):
            pyomo_con = reverse_con_map[cplex_con]
            slack[pyomo_con] = quadratic_vals[i]

    def load_duals(self, cons_to_load=None):
        """
        Load the duals into the 'dual' suffix. The 'dual' suffix must live on the parent model.

        Parameters
        ----------
        cons_to_load: list of Constraint
        """
        self._load_duals(cons_to_load)

    def load_rc(self, vars_to_load):
        """
        Load the reduced costs into the 'rc' suffix. The 'rc' suffix must live on the parent model.

        Parameters
        ----------
        vars_to_load: list of Var
        """
        self._load_rc(vars_to_load)

    def load_slacks(self, cons_to_load=None):
        """
        Load the values of the slack variables into the 'slack' suffix. The 'slack' suffix must live on the parent
        model.

        Parameters
        ----------
        cons_to_load: list of Constraint
        """
        self._load_slacks(cons_to_load)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
import glob

from pyomo.common.tempfiles import TempfileManager
from pyomo.solvers.plugins.solvers.ASL import ASL

logger = logging.getLogger('pyomo.contrib.gjh')


def readgjh(fname=None):
    """
    Build objective gradient and constraint Jacobian
    from gjh file written by the ASL gjh 'solver'.

    gjh solver may be called using pyomo and 'keepfiles=True'.
    Enable 'symbolic_solver_labels' as well to write
    .row and col file to get variable mappings.

    Parameters
    ----------
    fname : string, optional
        gjh file name. The default is None.

    Returns
    -------
    g : list
        Current objective gradient.
    J : list
        Current objective Jacobian.
    H : list
        Current objective Hessian.
    variableList : list
        Variables as defined by `*.col` file.
    constraintList : list
        Constraints as defined by `*.row` file.

    """
    if fname is None:
        files = list(glob.glob("*.gjh"))
        fname = files.pop(0)
        if len(files) > 1:
            print("WARNING: More than one gjh file in current directory")
            print("  Processing: %s\nIgnoring: %s" % (fname, '\n\t\t'.join(files)))

    with open(fname, "r") as f:
        line = "dummy_str"
        # Skip top lines until g value is reached
        while line != "param g :=\n":
            line = f.readline()

        line = f.readline()
        g = []
        # ; is the escape character
        while line[0] != ';':
            """
            When printed via ampl interface:
            ampl: display g;
            g [*] :=
                         1   0.204082
                         2   0.367347
                         3   0.44898
                         4   0.44898
                         5   0.244898
                         6  -0.173133
                         7  -0.173133
                         8  -0.0692532
                         9   0.0692532
                        10   0.346266
                        ;
            """
            index = int(line.split()[0]) - 1
            value = float(line.split()[1])
            g.append([index, value])
            line = f.readline()

        # Skip lines until J value is reached
        while line != "param J :=\n":
            line = f.readline()

        line = f.readline()
        J = []
        while line[0] != ';':
            """
            When printed via ampl interface:
            ampl: display J;
            J [*,*]
                :         1             2           3          4           5            6
                 :=
                1    -0.434327       0.784302     .          .           .          -0.399833
                2     2.22045e-16     .          1.46939     .           .          -0.831038
                3     0.979592        .           .         1.95918      .          -0.9596
                4     1.79592         .           .          .          2.12245     -0.692532
                5     0.979592        .           .          .           .           0
                6      .            -0.0640498   0.545265    .           .            .
                7      .             0.653061     .         1.14286      .            .
                8      .             1.63265      .          .          1.63265       .
                9      .             1.63265      .          .           .            .
                10     .              .          0.262481   0.262481     .            .
                11     .              .          1.14286     .          0.653061      .
                12     .              .          1.95918     .           .            .
                13     .              .           .         0.545265   -0.0640498     .
                14     .              .           .         1.95918      .            .
                15     .              .           .          .          1.63265       .
                16     .              .           .          .           .          -1

                :        7           8           9          10       :=
                1     0.399833     .           .          .
                2      .          0.831038     .          .
                3      .           .          0.9596      .
                4      .           .           .         0.692532
                6    -0.799667    0.799667     .          .
                7    -1.38506      .          1.38506     .
                8    -1.33278      .           .         1.33278
                9     0            .           .          .
                10     .         -0.9596      0.9596      .
                11     .         -1.38506      .         1.38506
                12     .          0            .          .
                13     .           .         -0.799667   0.799667
                14     .           .          0           .
                15     .           .           .         0
                16    1            .           .          .
                17   -1           1            .          .
                18     .         -1           1           .
                19     .           .         -1          1
            ;
            """
            if line[0] == '[':
                # Jacobian row index
                row = int(''.join(filter(str.isdigit, line))) - 1
                line = f.readline()

            column = int(line.split()[0]) - 1
            value = float(line.split()[1])
            J.append([row, column, value])
            line = f.readline()

        while line != "param H :=\n":
            line = f.readline()

        line = f.readline()
        H = []
        while line[0] != ';':
            """
            When printed via ampl interface:
            ampl: display H;
                H [*,*]
                :       1           2           3           4           5           6         :=
                1      .         0.25         .           .           .         -0.35348
                2     0.25        .          0.25         .           .         -0.212088
                3      .         0.25         .          0.25         .           .
                4      .          .          0.25         .          0.25         .
                5      .          .           .          0.25         .           .
                6    -0.35348   -0.212088     .           .           .         -0.0999584
                7     0.35348   -0.212088   -0.35348      .           .          0.0999584
                8      .         0.424176   -0.070696   -0.424176     .           .
                9      .          .          0.424176    0.070696   -0.424176     .
                10     .          .           .          0.35348     0.424176     .

                :        7            8           9          10        :=
                1     0.35348       .           .           .
                2    -0.212088     0.424176     .           .
                3    -0.35348     -0.070696    0.424176     .
                4      .          -0.424176    0.070696    0.35348
                5      .            .         -0.424176    0.424176
                6     0.0999584     .           .           .
                7    -0.299875     0.199917     .           .
                8     0.199917    -0.439817    0.2399       .
                9      .           0.2399     -0.439817    0.199917
                10     .            .          0.199917   -0.199917
                ;
            """
            if line[0] == '[':
                # Hessian row index
                row = int(''.join(filter(str.isdigit, line))) - 1
                line = f.readline()

            column = int(line.split()[0]) - 1
            value = float(line.split()[1])
            H.append([row, column, value])
            line = f.readline()

    with open(fname[:-3] + 'col', 'r') as f:
        data = f.read()
        variableList = data.split()

    with open(fname[:-3] + 'row', 'r') as f:
        data = f.read()
        constraintList = data.split()

    return g, J, H, variableList, constraintList


class GJHSolver(ASL):
    """
    An interface to the AMPL GJH "solver" for evaluating a model at a
    point.
    """

    def __init__(self, **kwds):
        kwds['type'] = 'gjh'
        kwds['symbolic_solver_labels'] = True
        super().__init__(**kwds)
        self.options.solver = 'gjh'
        self._metasolver = False

    # A hackish way to hold on to the model so that we can parse the
    # results.
    def _initialize_callbacks(self, model):
        self._model = model
        self._model._gjh_info = None
        super()._initialize_callbacks(model)

    def _presolve(self, *args, **kwds):
        super()._presolve(*args, **kwds)
        self._gjh_file = self._soln_file[:-3] + 'gjh'
        TempfileManager.add_tempfile(self._gjh_file, exists=False)

    def _postsolve(self):
        #
        # TODO: We should return the information using a better data
        # structure (ComponentMap?) so that the GJH solver does not need
        # to be called with symbolic_solver_labels=True
        #
        self._model._gjh_info = readgjh(self._gjh_file)
        self._model = None
        return super()._postsolve()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""This module provides functions for cut generation used across multiple
algorithms."""

from math import fabs
from pyomo.common.collections import ComponentSet
from pyomo.core import TransformationFactory, value, Constraint, Block


def _record_binary_value(var, var_value_is_one, var_value_is_zero, int_tol):
    val = value(var)
    if fabs(val - 1) <= int_tol:
        var_value_is_one.add(var)
    elif fabs(val) <= int_tol:
        var_value_is_zero.add(var)
    else:
        raise ValueError(
            'Binary %s = %s is not 0 or 1 within integer tolerance %s'
            % (var.name, val, int_tol)
        )


def add_no_good_cut(target_model_util_block, config):
    """Cut the current integer solution from the target model."""
    var_value_is_one = ComponentSet()
    var_value_is_zero = ComponentSet()
    for var in target_model_util_block.transformed_boolean_variable_list:
        _record_binary_value(
            var, var_value_is_one, var_value_is_zero, config.integer_tolerance
        )

    disjuncts = []
    if config.force_subproblem_nlp:
        # We need to also cut the solutions for the other discrete variables
        for var in target_model_util_block.discrete_variable_list:
            # This has possible duplicates with the above, but few enough that
            # we'll leave it for now.
            if var.is_binary():
                _record_binary_value(
                    var, var_value_is_one, var_value_is_zero, config.integer_tolerance
                )
            else:
                # It's integer. It still has to be in the no-good cut because
                # else the algorithm is wrong (We're cutting more than just this
                # solution), so we're going to do this as a Disjunction for
                # now. I think this is the *wrong* way to model if you are using
                # GDPopt, so if it's a little inefficient then... We'll deal
                # with that if it becomes an issue.
                val = round(value(var), 0)
                less = var <= val - 1
                more = var >= val + 1
                disjuncts.extend([less, more])

    # It shouldn't be possible to get here unless there's a solution to be cut.
    assert (var_value_is_one or var_value_is_zero) or len(disjuncts) == 0

    int_cut = (
        sum(1 - v for v in var_value_is_one) + sum(v for v in var_value_is_zero)
    ) >= 1

    if len(disjuncts) > 0:
        idx = len(target_model_util_block.no_good_disjunctions)
        target_model_util_block.no_good_disjunctions[idx] = [
            [disj] for disj in disjuncts
        ] + [[int_cut]]
        config.logger.debug(
            'Adding no-good disjunction: %s'
            % _disjunction_to_str(target_model_util_block.no_good_disjunctions[idx])
        )
        # transform it
        TransformationFactory(config.discrete_problem_transformation).apply_to(
            target_model_util_block,
            targets=[target_model_util_block.no_good_disjunctions[idx]],
        )
    else:
        config.logger.debug('Adding no-good cut: %s' % int_cut)
        # Exclude the current binary combination
        target_model_util_block.no_good_cuts.add(expr=int_cut)


def _disjunction_to_str(disjunction):
    pretty = []
    for disjunct in disjunction.disjuncts:
        exprs = []
        for cons in disjunct.component_data_objects(
            Constraint, active=True, descend_into=Block
        ):
            exprs.append(str(cons.expr))
        pretty.append("[%s]" % ", ".join(exprs))
    return " v ".join(pretty)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.collections import ComponentMap, ComponentSet
import pyomo.core.expr as _expr
from pyomo.core.expr.visitor import ExpressionValueVisitor, nonpyomo_leaf_types
from pyomo.core.expr.numvalue import value, is_constant
from pyomo.core.expr import exp, log, sin, cos
import math

"""
The purpose of this file is to perform symbolic differentiation and 
first order automatic differentiation directly with pyomo 
expressions. This is certainly not as efficient as doing AD in C or 
C++, but it avoids the translation from pyomo expressions to a form 
where AD can be performed efficiently. The only functions that are 
meant to be used by users are reverse_ad and reverse_sd. First, 
values are propagated from the leaves to each node in the tree with 
the LeafToRoot visitors. Then derivative values are propagated from 
the root to the leaves with the RootToLeaf visitors.
"""


class DifferentiationException(Exception):
    pass


def _diff_ProductExpression(node, val_dict, der_dict):
    """

    Parameters
    ----------
    node: pyomo.core.expr.numeric_expr.ProductExpression
    val_dict: ComponentMap
    der_dict: ComponentMap
    """
    assert len(node.args) == 2
    arg1, arg2 = node.args
    der = der_dict[node]
    der_dict[arg1] += der * val_dict[arg2]
    der_dict[arg2] += der * val_dict[arg1]


def _diff_SumExpression(node, val_dict, der_dict):
    """

    Parameters
    ----------
    node: pyomo.core.expr.numeric_expr.SumExpression
    val_dict: ComponentMap
    der_dict: ComponentMap
    """
    der = der_dict[node]
    for arg in node.args:
        der_dict[arg] += der


def _diff_PowExpression(node, val_dict, der_dict):
    """

    Parameters
    ----------
    node: pyomo.core.expr.numeric_expr.PowExpression
    val_dict: ComponentMap
    der_dict: ComponentMap
    """
    assert len(node.args) == 2
    arg1, arg2 = node.args
    der = der_dict[node]
    val1 = val_dict[arg1]
    val2 = val_dict[arg2]
    der_dict[arg1] += der * val2 * val1 ** (val2 - 1)
    if arg2.__class__ not in nonpyomo_leaf_types:
        der_dict[arg2] += der * val1**val2 * log(val1)


def _diff_DivisionExpression(node, val_dict, der_dict):
    """

    Parameters
    ----------
    node: pyomo.core.expr.numeric_expr.DivisionExpression
    val_dict: ComponentMap
    der_dict: ComponentMap
    """
    assert len(node.args) == 2
    num = node.args[0]
    den = node.args[1]
    der = der_dict[node]
    der_dict[num] += der * (1 / val_dict[den])
    der_dict[den] -= der * val_dict[num] / val_dict[den] ** 2


def _diff_NegationExpression(node, val_dict, der_dict):
    """

    Parameters
    ----------
    node: pyomo.core.expr.numeric_expr.UnaryFunctionExpression
    val_dict: ComponentMap
    der_dict: ComponentMap
    """
    assert len(node.args) == 1
    arg = node.args[0]
    der = der_dict[node]
    der_dict[arg] -= der


def _diff_exp(node, val_dict, der_dict):
    """

    Parameters
    ----------
    node: pyomo.core.expr.numeric_expr.UnaryFunctionExpression
    val_dict: ComponentMap
    der_dict: ComponentMap
    """
    assert len(node.args) == 1
    arg = node.args[0]
    der = der_dict[node]
    der_dict[arg] += der * exp(val_dict[arg])


def _diff_log(node, val_dict, der_dict):
    """

    Parameters
    ----------
    node: pyomo.core.expr.numeric_expr.UnaryFunctionExpression
    val_dict: ComponentMap
    der_dict: ComponentMap
    """
    assert len(node.args) == 1
    arg = node.args[0]
    der = der_dict[node]
    der_dict[arg] += der / val_dict[arg]


def _diff_log10(node, val_dict, der_dict):
    """

    Parameters
    ----------
    node: pyomo.core.expr.numeric_expr.UnaryFunctionExpression
    val_dict: ComponentMap
    der_dict: ComponentMap
    """
    assert len(node.args) == 1
    arg = node.args[0]
    der = der_dict[node]
    der_dict[arg] += der * math.log10(math.exp(1)) / val_dict[arg]


def _diff_sin(node, val_dict, der_dict):
    """

    Parameters
    ----------
    node: pyomo.core.expr.numeric_expr.UnaryFunctionExpression
    val_dict: ComponentMap
    der_dict: ComponentMap
    """
    assert len(node.args) == 1
    arg = node.args[0]
    der = der_dict[node]
    der_dict[arg] += der * cos(val_dict[arg])


def _diff_cos(node, val_dict, der_dict):
    """

    Parameters
    ----------
    node: pyomo.core.expr.numeric_expr.UnaryFunctionExpression
    val_dict: ComponentMap
    der_dict: ComponentMap
    """
    assert len(node.args) == 1
    arg = node.args[0]
    der = der_dict[node]
    der_dict[arg] -= der * sin(val_dict[arg])


def _diff_tan(node, val_dict, der_dict):
    """

    Parameters
    ----------
    node: pyomo.core.expr.numeric_expr.UnaryFunctionExpression
    val_dict: ComponentMap
    der_dict: ComponentMap
    """
    assert len(node.args) == 1
    arg = node.args[0]
    der = der_dict[node]
    der_dict[arg] += der / (cos(val_dict[arg]) ** 2)


def _diff_asin(node, val_dict, der_dict):
    """

    Parameters
    ----------
    node: pyomo.core.expr.numeric_expr.UnaryFunctionExpression
    val_dict: ComponentMap
    der_dict: ComponentMap
    """
    assert len(node.args) == 1
    arg = node.args[0]
    der = der_dict[node]
    der_dict[arg] += der / (1 - val_dict[arg] ** 2) ** 0.5


def _diff_acos(node, val_dict, der_dict):
    """

    Parameters
    ----------
    node: pyomo.core.expr.numeric_expr.UnaryFunctionExpression
    val_dict: ComponentMap
    der_dict: ComponentMap
    """
    assert len(node.args) == 1
    arg = node.args[0]
    der = der_dict[node]
    der_dict[arg] -= der / (1 - val_dict[arg] ** 2) ** 0.5


def _diff_atan(node, val_dict, der_dict):
    """

    Parameters
    ----------
    node: pyomo.core.expr.numeric_expr.UnaryFunctionExpression
    val_dict: ComponentMap
    der_dict: ComponentMap
    """
    assert len(node.args) == 1
    arg = node.args[0]
    der = der_dict[node]
    der_dict[arg] += der / (1 + val_dict[arg] ** 2)


def _diff_sqrt(node, val_dict, der_dict):
    """
    Reverse automatic differentiation on the square root function.
    Implementation copied from power function, with fixed exponent.

    Parameters
    ----------
    node: pyomo.core.expr.numeric_expr.UnaryFunctionExpression
    val_dict: ComponentMap
    der_dict: ComponentMap
    """
    assert len(node.args) == 1
    arg = node.args[0]
    der = der_dict[node]
    der_dict[arg] += der * 0.5 * val_dict[arg] ** (-0.5)


def _diff_abs(node, val_dict, der_dict):
    """
    Reverse automatic differentiation on the abs function.
    This will raise an exception at 0.

    Parameters
    ----------
    node: pyomo.core.expr.numeric_expr.UnaryFunctionExpression
    val_dict: ComponentMap
    der_dict: ComponentMap
    """
    assert len(node.args) == 1
    arg = node.args[0]
    der = der_dict[node]
    val = val_dict[arg]
    if is_constant(val) and val == 0:
        raise DifferentiationException('Cannot differentiate abs(x) at x=0')
    der_dict[arg] += der * val / abs(val)


_unary_map = dict()
_unary_map['exp'] = _diff_exp
_unary_map['log'] = _diff_log
_unary_map['log10'] = _diff_log10
_unary_map['sin'] = _diff_sin
_unary_map['cos'] = _diff_cos
_unary_map['tan'] = _diff_tan
_unary_map['asin'] = _diff_asin
_unary_map['acos'] = _diff_acos
_unary_map['atan'] = _diff_atan
_unary_map['sqrt'] = _diff_sqrt
_unary_map['abs'] = _diff_abs


def _diff_UnaryFunctionExpression(node, val_dict, der_dict):
    """

    Parameters
    ----------
    node: pyomo.core.expr.numeric_expr.UnaryFunctionExpression
    val_dict: ComponentMap
    der_dict: ComponentMap
    """
    if node.getname() in _unary_map:
        _unary_map[node.getname()](node, val_dict, der_dict)
    else:
        raise DifferentiationException(
            'Unsupported expression type for differentiation: {0}'.format(type(node))
        )


def _diff_GeneralExpression(node, val_dict, der_dict):
    """
    Reverse automatic differentiation for named expressions.

    Parameters
    ----------
    node: The named expression
    val_dict: ComponentMap
    der_dict: ComponentMap
    """
    der_dict[node.arg(0)] += der_dict[node]


def _diff_ExternalFunctionExpression(node, val_dict, der_dict):
    """

    Parameters
    ----------
    node: pyomo.core.expr.numeric_expr.ExternalFunctionExpression
    val_dict: ComponentMap
    der_dict: ComponentMap
    """
    der = der_dict[node]
    vals = tuple(val_dict[i] for i in node.args)
    derivs = node._fcn.evaluate_fgh(vals, fgh=1)[1]
    for ndx, arg in enumerate(node.args):
        der_dict[arg] += der * derivs[ndx]


_diff_map = dict()
_diff_map[_expr.ProductExpression] = _diff_ProductExpression
_diff_map[_expr.DivisionExpression] = _diff_DivisionExpression
_diff_map[_expr.PowExpression] = _diff_PowExpression
_diff_map[_expr.SumExpression] = _diff_SumExpression
_diff_map[_expr.MonomialTermExpression] = _diff_ProductExpression
_diff_map[_expr.NegationExpression] = _diff_NegationExpression
_diff_map[_expr.UnaryFunctionExpression] = _diff_UnaryFunctionExpression
_diff_map[_expr.ExternalFunctionExpression] = _diff_ExternalFunctionExpression
_diff_map[_expr.LinearExpression] = _diff_SumExpression
_diff_map[_expr.AbsExpression] = _diff_abs

_diff_map[_expr.NPV_ProductExpression] = _diff_ProductExpression
_diff_map[_expr.NPV_DivisionExpression] = _diff_DivisionExpression
_diff_map[_expr.NPV_PowExpression] = _diff_PowExpression
_diff_map[_expr.NPV_SumExpression] = _diff_SumExpression
_diff_map[_expr.NPV_NegationExpression] = _diff_NegationExpression
_diff_map[_expr.NPV_UnaryFunctionExpression] = _diff_UnaryFunctionExpression
_diff_map[_expr.NPV_ExternalFunctionExpression] = _diff_ExternalFunctionExpression
_diff_map[_expr.NPV_AbsExpression] = _diff_abs


def _symbolic_value(x):
    return x


def _numeric_apply_operation(node, values):
    return node._apply_operation(values)


def _symbolic_apply_operation(node, values):
    return node


class _LeafToRootVisitor(ExpressionValueVisitor):
    def __init__(self, val_dict, der_dict, expr_list, numeric=True):
        """
        Parameters
        ----------
        val_dict: ComponentMap
        der_dict: ComponentMap
        """
        self.val_dict = val_dict
        self.der_dict = der_dict
        self.expr_list = expr_list
        assert len(self.expr_list) == 0
        assert len(self.val_dict) == 0
        assert len(self.der_dict) == 0
        if numeric:
            self.value_func = value
            self.operation_func = _numeric_apply_operation
        else:
            self.value_func = _symbolic_value
            self.operation_func = _symbolic_apply_operation

    def visit(self, node, values):
        self.val_dict[node] = self.operation_func(node, values)
        self.der_dict[node] = 0
        self.expr_list.append(node)
        return self.val_dict[node]

    def visiting_potential_leaf(self, node):
        if node in self.val_dict:
            return True, self.val_dict[node]

        if node.__class__ in nonpyomo_leaf_types:
            self.val_dict[node] = node
            self.der_dict[node] = 0
            return True, node

        if not node.is_expression_type():
            val = self.value_func(node)
            self.val_dict[node] = val
            self.der_dict[node] = 0
            return True, val

        return False, None


def _reverse_diff_helper(expr, numeric=True):
    val_dict = ComponentMap()
    der_dict = ComponentMap()
    expr_list = list()

    visitorA = _LeafToRootVisitor(val_dict, der_dict, expr_list, numeric=numeric)
    visitorA.dfs_postorder_stack(expr)

    der_dict[expr] = 1
    for e in reversed(expr_list):
        if e.__class__ in _diff_map:
            _diff_map[e.__class__](e, val_dict, der_dict)
        elif e.is_named_expression_type():
            _diff_GeneralExpression(e, val_dict, der_dict)
        else:
            raise DifferentiationException(
                'Unsupported expression type for differentiation: {0}'.format(type(e))
            )

    return der_dict


def reverse_ad(expr):
    """
    First order reverse automatic differentiation

    Parameters
    ----------
    expr: pyomo.core.expr.numeric_expr.NumericExpression
        expression to differentiate

    Returns
    -------
    ComponentMap
        component_map mapping variables to derivatives with respect
        to the corresponding variable
    """
    return _reverse_diff_helper(expr, True)


def reverse_sd(expr):
    """
    First order reverse symbolic differentiation

    Parameters
    ----------
    expr: pyomo.core.expr.numeric_expr.NumericExpression
        expression to differentiate

    Returns
    -------
    ComponentMap
        component_map mapping variables to derivatives with respect
        to the corresponding variable
    """
    return _reverse_diff_helper(expr, False)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import json
import sys
import argparse

from pyomo.common.collections import Bunch
from pyomo.opt import guess_format
from pyomo.scripting.pyomo_parser import add_subparser, CustomHelpFormatter
from pyomo.scripting.solve_config import Default_Config


def create_parser(parser=None):
    #
    # Setup command-line options.
    #
    if parser is None:
        parser = argparse.ArgumentParser(
            usage='%(prog)s [options] <model_or_config_file> [<data_files>]'
        )
    parser.add_argument(
        '--output',
        action='store',
        dest='filename',
        help="Output file name. This option is required unless the file name is specified in a       configuration file.",
        default=None,
    )
    parser.add_argument(
        '--format', action='store', dest='format', help="Output format", default=None
    )
    parser.add_argument(
        '--generate-config-template', action='store', dest='template', default=None
    )
    return parser


def run_convert(options=Bunch(), parser=None):
    from pyomo.scripting.convert import convert, convert_dakota

    if options.model.save_format is None and options.model.save_file:
        options.model.save_format = options.model.save_file.split('.')[-1]
    #
    _format = guess_format(options.model.save_format)

    if options.model.save_format == 'dakota':
        return convert_dakota(options, parser)
    elif _format is None:
        if options.model.save_format is None:
            raise RuntimeError("Unspecified target conversion format!")
        else:
            raise RuntimeError(
                "Unrecognized target conversion format (%s)!"
                % (options.model.save_format,)
            )
    else:
        return convert(options, parser, _format)


def convert_exec(args, unparsed):
    #
    import pyomo.scripting.util

    #
    # Generate a template file
    #
    if not args.template is None:
        config, blocks = Default_Config().config_block(True)
        OUTPUT = open(args.template, 'w')
        if args.template.endswith('json'):
            OUTPUT.write(json.dumps(config.value(), indent=2))
        else:
            OUTPUT.write(config.generate_yaml_template())
        OUTPUT.close()
        print("  Created template file '%s'" % args.template)
        sys.exit(0)
    #
    save_filename = getattr(args, 'filename', None)
    if save_filename is None:
        save_format = getattr(args, 'format', None)
        if not save_format is None:
            save_filename = 'unknown.' + save_format
    if save_filename is None:
        #
        # Get configuration values if no model file has been specified
        #
        try:
            val = pyomo.scripting.util.get_config_values(unparsed[-1])
        except IndexError:
            val = None
        except IOError:
            val = None
        #
        # Try to get the model file
        #
        if not val is None:
            try:
                save_filename = val['model']['save file']
            except:
                pass
            if save_filename is None:
                try:
                    save_filename = 'unknown.' + str(val['model']['save format'])
                except:
                    pass
    #
    # Help options or no filename have been specified, then print help ...
    #
    if save_filename is None or '-h' in unparsed or '--help' in unparsed:
        if not ('-h' in unparsed or '--help' in unparsed):
            print("ERROR: No output file or format specified!")
            print("")
        config, blocks = Default_Config().config_block()
        parser = create_temporary_parser(output=True, generate=True)
        config.initialize_argparse(parser)
        parser.parse_args(args=unparsed + ['-h'])
        sys.exit(1)
    #
    # Parse previously unparsed options
    #
    config, blocks = Default_Config().config_block()
    _parser = create_temporary_parser()
    config.initialize_argparse(_parser)
    _options = _parser.parse_args(args=unparsed)
    #
    # Import the parsed values into the config block, then
    # create an Options object
    #
    config.import_argparse(_options)
    config.model.save_file = getattr(args, 'filename', None)
    config.model.save_format = getattr(args, 'format', None)
    if _options.model_or_config_file.endswith('.py'):
        config.model.filename = _options.model_or_config_file
        config.data.files = _options.data_files
    else:
        val = pyomo.scripting.util.get_config_values(_options.model_or_config_file)
        config.set_value(val)
    #
    # Note that we pass-in pre-parsed options.  The run_command()
    # function knows to not perform a parse, but instead to simply
    # used these parsed values.
    #
    return pyomo.scripting.util.run_command(
        command=run_convert, parser=convert_parser, options=config, name='convert'
    )


#
# Add a subparser for the pyomo command
#
convert_parser = create_parser(
    add_subparser(
        'convert',
        func=convert_exec,
        help='Convert a Pyomo model to another format',
        add_help=False,
        description='This pyomo subcommand is used to create a new model file in a specified format from a Pyomo model.',
    )
)


def create_temporary_parser(output=False, generate=False):
    #
    # We create a dummy parser and subparser, to help make the 'help'
    # output seem sensible.
    #
    parser = argparse.ArgumentParser(formatter_class=CustomHelpFormatter)
    _subparsers = parser.add_subparsers()
    _parser = _subparsers.add_parser('convert')
    _parser.formatter_class = CustomHelpFormatter
    if generate:
        #
        # Adding documentation about the two options that are
        # defined in the initial parser.
        #
        _parser.add_argument(
            '--generate-config-template',
            action='store',
            dest='template',
            default=None,
            help='Create a configuration template file in YAML or JSON and exit.',
        )
    if output:
        parser.add_argument(
            '--output',
            action='store',
            dest='filename',
            help="Output file name. This option is required unless the file name is specified in a       configuration file.",
            default=None,
        )
        parser.add_argument(
            '--format',
            action='store',
            dest='format',
            help="Output format",
            default=None,
        )
        _parser.usage = '%(prog)s [options] <model_or_config_file> [<data_files>]'
        _parser.epilog = """
Description:

The 'pyomo convert' subcommand converts a Pyomo model into a specified
format.  The --output option is used to specify an output file.
If only the --format option is specified, then the output filename is
unknown.<format>.  The standard steps executed by this subcommand are:

  - Apply pre-solve operations (e.g. import Python packages)
  - Create the model
  - Apply model transformations


This subcommand can be executed with or without a configuration file.
For example:

  pyomo convert --output=model.lp model.py model.dat

This creates the file 'model.lp' with format 'lp'.


A yaml or json configuration file can also be used to specify conversion
options.  For example:

  pyomo convert config.yaml

No other command-line options are required!

A template configuration file can be generated with the
'--generate-config-template' option:

  pyomo convert --generate-config-template=template.yaml
  pyomo convert --generate-config-template=template.json

Note that yaml template files contain comments that describe the
configuration options.  Also, configuration files will generally support
more configuration options than are available with command-line options.

"""
    #
    _parser.add_argument(
        '--output',
        action='store',
        dest='filename',
        help="Output file name. This option is required unless the file name is specified in a       configuration file.",
        default=None,
    )
    _parser.add_argument(
        '--format', action='store', dest='format', help="Output format", default=None
    )
    _parser.add_argument(
        'model_or_config_file',
        action='store',
        nargs='?',
        default='',
        help="A Python module that defines a Pyomo model, or a configuration file that defines options for 'pyomo convert' (in either YAML or JSON format)",
    )
    _parser.add_argument(
        'data_files',
        action='store',
        nargs='*',
        default=[],
        help='Pyomo data files that defined data used to initialize the model (specified in the first argument)',
    )
    #
    return _parser



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from io import StringIO
import pyomo.common.unittest as unittest
from pyomo.contrib.cp.interval_var import IntervalVar
from pyomo.contrib.cp.sequence_var import SequenceVar, IndexedSequenceVar
from pyomo.environ import ConcreteModel, Set


class TestScalarSequenceVar(unittest.TestCase):
    def test_initialize_with_no_data(self):
        m = ConcreteModel()
        m.i = SequenceVar()

        self.assertIsInstance(m.i, SequenceVar)
        self.assertIsInstance(m.i.interval_vars, list)
        self.assertEqual(len(m.i.interval_vars), 0)

        m.iv1 = IntervalVar()
        m.iv2 = IntervalVar()
        m.i.set_value(expr=[m.iv1, m.iv2])

        self.assertIsInstance(m.i.interval_vars, list)
        self.assertEqual(len(m.i.interval_vars), 2)
        self.assertIs(m.i.interval_vars[0], m.iv1)
        self.assertIs(m.i.interval_vars[1], m.iv2)

    def get_model(self):
        m = ConcreteModel()
        m.S = Set(initialize=range(3))
        m.i = IntervalVar(m.S, start=(0, 5))
        m.seq = SequenceVar(expr=[m.i[j] for j in m.S])

        return m

    def test_initialize_with_expr(self):
        m = self.get_model()
        self.assertEqual(len(m.seq.interval_vars), 3)
        for j in m.S:
            self.assertIs(m.seq.interval_vars[j], m.i[j])

    def test_pprint(self):
        m = self.get_model()
        buf = StringIO()
        m.seq.pprint(ostream=buf)
        self.assertEqual(
            buf.getvalue().strip(),
            """
seq : Size=1, Index=None
    Key  : IntervalVars
    None : [i[0], i[1], i[2]]
            """.strip(),
        )

    def test_interval_vars_not_a_list(self):
        m = self.get_model()

        with self.assertRaisesRegex(
            ValueError,
            "'expr' for SequenceVar must be a list of IntervalVars. "
            "Encountered type '<class 'int'>' constructing 'seq2'",
        ):
            m.seq2 = SequenceVar(expr=1)

    def test_interval_vars_list_includes_things_that_are_not_interval_vars(self):
        m = self.get_model()

        with self.assertRaisesRegex(
            ValueError,
            "The SequenceVar 'expr' argument must be a list of "
            "IntervalVars. The 'expr' for SequenceVar 'seq2' included "
            "an object of type '<class 'int'>'",
        ):
            m.seq2 = SequenceVar(expr=m.i)


class TestIndexedSequenceVar(unittest.TestCase):
    def test_initialize_with_not_data(self):
        m = ConcreteModel()
        m.i = SequenceVar([1, 2])

        self.assertIsInstance(m.i, IndexedSequenceVar)
        for j in [1, 2]:
            self.assertIsInstance(m.i[j].interval_vars, list)
            self.assertEqual(len(m.i[j].interval_vars), 0)

        m.iv = IntervalVar()
        m.iv2 = IntervalVar([0, 1])
        m.i[2] = [m.iv] + [m.iv2[i] for i in [0, 1]]

        self.assertEqual(len(m.i[2].interval_vars), 3)
        self.assertEqual(len(m.i[1].interval_vars), 0)
        self.assertIs(m.i[2].interval_vars[0], m.iv)
        for i in [0, 1]:
            self.assertIs(m.i[2].interval_vars[i + 1], m.iv2[i])

    def make_model(self):
        m = ConcreteModel()
        m.alphabetic = Set(initialize=['a', 'b'])
        m.numeric = Set(initialize=[1, 2])
        m.i = IntervalVar(m.alphabetic, m.numeric)

        def the_rule(m, j):
            return [m.i[j, k] for k in m.numeric]

        m.seq = SequenceVar(m.alphabetic, rule=the_rule)

        return m

    def test_initialize_with_rule(self):
        m = self.make_model()

        self.assertIsInstance(m.seq, IndexedSequenceVar)
        self.assertEqual(len(m.seq), 2)
        for j in m.alphabetic:
            self.assertTrue(j in m.seq)
            self.assertEqual(len(m.seq[j].interval_vars), 2)
            for k in m.numeric:
                self.assertIs(m.seq[j].interval_vars[k - 1], m.i[j, k])

    def test_pprint(self):
        m = self.make_model()
        m.seq.pprint()

        buf = StringIO()
        m.seq.pprint(ostream=buf)
        self.assertEqual(
            buf.getvalue().strip(),
            """
seq : Size=2, Index=alphabetic
    Key : IntervalVars
      a : [i[a,1], i[a,2]]
      b : [i[b,1], i[b,2]]""".strip(),
        )

    def test_multidimensional_index(self):
        m = self.make_model()

        @m.SequenceVar(m.alphabetic, m.numeric)
        def s(m, i, j):
            return [m.i[i, j]]

        self.assertIsInstance(m.s, IndexedSequenceVar)
        self.assertEqual(len(m.s), 4)
        for i in m.alphabetic:
            for j in m.numeric:
                self.assertTrue((i, j) in m.s)
                self.assertEqual(len(m.s[i, j].interval_vars), 1)
                self.assertIs(m.s[i, j].interval_vars[0], m.i[i, j])



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
import sys

from importlib import import_module
from io import StringIO

from pyomo.common.log import LoggingIntercept


def force_load(module):
    if module in sys.modules:
        del sys.modules[module]
    return import_module(module)


class TestDeprecatedModules(unittest.TestCase):
    def test_rangeset(self):
        log = StringIO()
        with LoggingIntercept(log):
            from pyomo.core.base.set import RangeSet
        self.assertEqual(log.getvalue(), "")

        log = StringIO()
        with LoggingIntercept(log, 'pyomo'):
            rs = force_load('pyomo.core.base.rangeset')
        self.assertIn(
            "The pyomo.core.base.rangeset module is deprecated.",
            log.getvalue().strip().replace('\n', ' '),
        )
        self.assertIs(RangeSet, rs.RangeSet)

        # Run this twice to implicitly test the force_load() implementation
        log = StringIO()
        with LoggingIntercept(log, 'pyomo'):
            rs = force_load('pyomo.core.base.rangeset')
        self.assertIn(
            "The pyomo.core.base.rangeset module is deprecated.",
            log.getvalue().strip().replace('\n', ' '),
        )
        self.assertIs(RangeSet, rs.RangeSet)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
import logging

import pyomo.common.unittest as unittest
import pyomo.environ as pyo
import pyomo.solvers.plugins.solvers.xpress_direct as xpd

from pyomo.common.log import LoggingIntercept
from pyomo.core.expr.taylor_series import taylor_series_expansion
from pyomo.opt.results.solver import TerminationCondition, SolverStatus
from pyomo.solvers.plugins.solvers.xpress_persistent import XpressPersistent

xpress_available = pyo.SolverFactory('xpress_persistent').available(False)


@unittest.skipIf(not xpress_available, "xpress is not available")
@unittest.pytest.mark.solver("xpress_persistent")
class TestXpressPersistent(unittest.TestCase):
    def test_basics(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(bounds=(-10, 10))
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.x**2 + m.y**2)
        m.c1 = pyo.Constraint(expr=m.y >= 2 * m.x + 1)

        opt = pyo.SolverFactory('xpress_persistent')
        opt.set_instance(m)

        self.assertEqual(opt.get_xpress_attribute('cols'), 2)
        self.assertEqual(opt.get_xpress_attribute('rows'), 1)

        res = opt.solve()
        self.assertAlmostEqual(m.x.value, -0.4, delta=1e-6)
        self.assertAlmostEqual(m.y.value, 0.2, delta=1e-6)

        opt.load_duals()
        self.assertEqual(len(m.dual), 1)
        self.assertAlmostEqual(m.dual[m.c1], -0.4, delta=1e-6)
        del m.dual

        opt.load_rc()
        self.assertEqual(len(m.rc), 2)
        self.assertAlmostEqual(m.rc[m.x], 0, delta=1e-8)
        self.assertAlmostEqual(m.rc[m.y], 0, delta=1e-8)
        del m.rc

        opt.load_slacks()
        self.assertEqual(len(m.slack), 1)
        self.assertAlmostEqual(m.slack[m.c1], 0, delta=1e-6)
        del m.slack

        m.c2 = pyo.Constraint(expr=m.y >= -m.x + 1)
        opt.add_constraint(m.c2)
        self.assertEqual(opt.get_xpress_attribute('cols'), 2)
        self.assertEqual(opt.get_xpress_attribute('rows'), 2)

        res = opt.solve(save_results=False, load_solutions=False)
        self.assertAlmostEqual(m.x.value, -0.4, delta=1e-6)
        self.assertAlmostEqual(m.y.value, 0.2, delta=1e-6)
        opt.load_vars()
        self.assertAlmostEqual(m.x.value, 0, delta=2.5e-6)
        self.assertAlmostEqual(m.y.value, 1, delta=2.5e-6)

        opt.remove_constraint(m.c2)
        m.del_component(m.c2)
        self.assertEqual(opt.get_xpress_attribute('cols'), 2)
        self.assertEqual(opt.get_xpress_attribute('rows'), 1)

        self.assertEqual(opt.get_xpress_control('feastol'), 1e-6)
        res = opt.solve(options={'feastol': '1e-7'})
        self.assertEqual(opt.get_xpress_control('feastol'), 1e-7)
        self.assertAlmostEqual(m.x.value, -0.4, delta=1e-6)
        self.assertAlmostEqual(m.y.value, 0.2, delta=1e-6)

        m.x.setlb(-5)
        m.x.setub(5)
        opt.update_var(m.x)
        lb = opt._get_lb(m.x)
        ub = opt._get_ub(m.x)
        self.assertEqual(lb, -5)
        self.assertEqual(ub, 5)

        m.x.fix(0)
        opt.update_var(m.x)
        lb = opt._get_lb(m.x)
        ub = opt._get_ub(m.x)
        self.assertEqual(lb, 0)
        self.assertEqual(ub, 0)

        m.x.unfix()
        opt.update_var(m.x)
        lb = opt._get_lb(m.x)
        ub = opt._get_ub(m.x)
        self.assertEqual(lb, -5)
        self.assertEqual(ub, 5)

        m.c2 = pyo.Constraint(expr=m.y >= m.x**2)
        opt.add_constraint(m.c2)
        self.assertEqual(opt.get_xpress_attribute('cols'), 2)
        self.assertEqual(opt.get_xpress_attribute('rows'), 2)

        opt.remove_constraint(m.c2)
        m.del_component(m.c2)
        self.assertEqual(opt.get_xpress_attribute('cols'), 2)
        self.assertEqual(opt.get_xpress_attribute('rows'), 1)

        m.z = pyo.Var()
        opt.add_var(m.z)
        self.assertEqual(opt.get_xpress_attribute('cols'), 3)
        opt.remove_var(m.z)
        del m.z
        self.assertEqual(opt.get_xpress_attribute('cols'), 2)

    def test_vartype_change(self):
        # test for issue #3565
        m = pyo.ConcreteModel()
        m.x = pyo.Var(bounds=(0, 1))
        m.o = pyo.Objective(expr=m.x)

        opt = pyo.SolverFactory('xpress_persistent')
        opt.set_instance(m)

        m.x.fix(1)
        opt.update_var(m.x)

        lb = opt._get_lb(m.x)
        self.assertEqual(lb, 1)

        m.x.domain = pyo.Binary
        opt.update_var(m.x)

        lb = opt._get_lb(m.x)
        self.assertEqual(lb, 1)

    def test_add_remove_qconstraint(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.z = pyo.Var()
        m.obj = pyo.Objective(expr=m.z)
        m.c1 = pyo.Constraint(expr=m.z >= m.x**2 + m.y**2)

        opt = pyo.SolverFactory('xpress_persistent')
        opt.set_instance(m)
        self.assertEqual(opt.get_xpress_attribute('rows'), 1)

        opt.remove_constraint(m.c1)
        self.assertEqual(opt.get_xpress_attribute('rows'), 0)

        opt.add_constraint(m.c1)
        self.assertEqual(opt.get_xpress_attribute('rows'), 1)

    def test_add_remove_lconstraint(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.z = pyo.Var()
        m.obj = pyo.Objective(expr=m.z)
        m.c2 = pyo.Constraint(expr=m.x + m.y == 1)

        opt = pyo.SolverFactory('xpress_persistent')
        opt.set_instance(m)
        self.assertEqual(opt.get_xpress_attribute('rows'), 1)

        opt.remove_constraint(m.c2)
        self.assertEqual(opt.get_xpress_attribute('rows'), 0)

        opt.add_constraint(m.c2)
        self.assertEqual(opt.get_xpress_attribute('rows'), 1)

    def test_add_remove_sosconstraint(self):
        m = pyo.ConcreteModel()
        m.a = pyo.Set(initialize=[1, 2, 3], ordered=True)
        m.x = pyo.Var(m.a, within=pyo.Binary)
        m.y = pyo.Var(within=pyo.Binary)
        m.obj = pyo.Objective(expr=m.y)
        m.c1 = pyo.SOSConstraint(var=m.x, sos=1)

        opt = pyo.SolverFactory('xpress_persistent')
        opt.set_instance(m)
        self.assertEqual(opt.get_xpress_attribute('sets'), 1)

        opt.remove_sos_constraint(m.c1)
        self.assertEqual(opt.get_xpress_attribute('sets'), 0)

        opt.add_sos_constraint(m.c1)
        self.assertEqual(opt.get_xpress_attribute('sets'), 1)

    def test_add_remove_sosconstraint2(self):
        m = pyo.ConcreteModel()
        m.a = pyo.Set(initialize=[1, 2, 3], ordered=True)
        m.x = pyo.Var(m.a, within=pyo.Binary)
        m.y = pyo.Var(within=pyo.Binary)
        m.obj = pyo.Objective(expr=m.y)
        m.c1 = pyo.SOSConstraint(var=m.x, sos=1)

        opt = pyo.SolverFactory('xpress_persistent')
        opt.set_instance(m)
        self.assertEqual(opt.get_xpress_attribute('sets'), 1)
        m.c2 = pyo.SOSConstraint(var=m.x, sos=2)
        opt.add_sos_constraint(m.c2)
        self.assertEqual(opt.get_xpress_attribute('sets'), 2)
        opt.remove_sos_constraint(m.c2)
        self.assertEqual(opt.get_xpress_attribute('sets'), 1)

    def test_add_remove_var(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()

        opt = pyo.SolverFactory('xpress_persistent')
        opt.set_instance(m)
        self.assertEqual(opt.get_xpress_attribute('cols'), 2)

        opt.remove_var(m.x)
        self.assertEqual(opt.get_xpress_attribute('cols'), 1)

        opt.add_var(m.x)
        self.assertEqual(opt.get_xpress_attribute('cols'), 2)

        opt.remove_var(m.x)
        opt.add_var(m.x)
        opt.remove_var(m.x)
        self.assertEqual(opt.get_xpress_attribute('cols'), 1)

    def test_add_column(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(within=pyo.NonNegativeReals)
        m.c = pyo.Constraint(expr=(0, m.x, 1))
        m.obj = pyo.Objective(expr=-m.x)

        opt = pyo.SolverFactory('xpress_persistent')
        opt.set_instance(m)
        opt.solve()
        self.assertAlmostEqual(m.x.value, 1)

        m.y = pyo.Var(within=pyo.NonNegativeReals)

        opt.add_column(m, m.y, -3, [m.c], [2])
        opt.solve()

        self.assertAlmostEqual(m.x.value, 0)
        self.assertAlmostEqual(m.y.value, 0.5)

    def test_add_column_exceptions(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.c = pyo.Constraint(expr=(0, m.x, 1))
        m.ci = pyo.Constraint([1, 2], rule=lambda m, i: (0, m.x, i + 1))
        m.cd = pyo.Constraint(expr=(0, -m.x, 1))
        m.cd.deactivate()
        m.obj = pyo.Objective(expr=-m.x)

        opt = pyo.SolverFactory('xpress_persistent')

        # set_instance not called
        self.assertRaises(RuntimeError, opt.add_column, m, m.x, 0, [m.c], [1])

        opt.set_instance(m)

        m2 = pyo.ConcreteModel()
        m2.y = pyo.Var()
        m2.c = pyo.Constraint(expr=(0, m.x, 1))

        # different model than attached to opt
        self.assertRaises(RuntimeError, opt.add_column, m2, m2.y, 0, [], [])
        # pyomo var attached to different model
        self.assertRaises(RuntimeError, opt.add_column, m, m2.y, 0, [], [])

        z = pyo.Var()
        # pyomo var floating
        self.assertRaises(RuntimeError, opt.add_column, m, z, -2, [m.c, z], [1])

        m.y = pyo.Var()
        # len(coefficients) == len(constraints)
        self.assertRaises(RuntimeError, opt.add_column, m, m.y, -2, [m.c], [1, 2])
        self.assertRaises(RuntimeError, opt.add_column, m, m.y, -2, [m.c, z], [1])

        # add indexed constraint
        self.assertRaises(AttributeError, opt.add_column, m, m.y, -2, [m.ci], [1])
        # add something not a ConstraintData
        self.assertRaises(AttributeError, opt.add_column, m, m.y, -2, [m.x], [1])

        # constraint not on solver model
        self.assertRaises(KeyError, opt.add_column, m, m.y, -2, [m2.c], [1])

        # inactive constraint
        self.assertRaises(KeyError, opt.add_column, m, m.y, -2, [m.cd], [1])

        opt.add_var(m.y)
        # var already in solver model
        self.assertRaises(RuntimeError, opt.add_column, m, m.y, -2, [m.c], [1])

    @unittest.skipIf(
        xpd.xpress_available and xpd.xpress.__version__ == '9.8.0',
        "Xpress 9.8 always runs global optimizer",
    )
    def test_nonconvexqp_locally_optimal(self):
        """Test non-convex QP for which xpress_direct should find a locally
        optimal solution."""
        m = pyo.ConcreteModel()
        m.x1 = pyo.Var()
        m.x2 = pyo.Var()
        m.x3 = pyo.Var()

        m.obj = pyo.Objective(rule=lambda m: 2 * m.x1 + m.x2 + m.x3, sense=pyo.minimize)
        m.equ1 = pyo.Constraint(rule=lambda m: m.x1 + m.x2 + m.x3 == 1)
        m.cone = pyo.Constraint(rule=lambda m: m.x2 * m.x2 + m.x3 * m.x3 <= m.x1 * m.x1)
        m.equ2 = pyo.Constraint(rule=lambda m: m.x1 >= 0)

        opt = pyo.SolverFactory('xpress_direct')
        opt.options['XSLP_SOLVER'] = 0
        # xpress 9.5.0 now defaults to trying (and failing) to solve this problem
        # using the global solver. This option forces the use of the local solver.
        opt.options['NLPSOLVER'] = 1

        results = opt.solve(m)
        self.assertEqual(results.solver.status, SolverStatus.ok)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.locallyOptimal
        )

        # Cannot test exact values since the may be different depending on
        # random effects. So just test all are non-zero.
        self.assertGreater(m.x1.value, 0.0)
        self.assertGreater(m.x2.value, 0.0)
        self.assertGreater(m.x3.value, 0.0)

    def test_nonconvexqp_infeasible(self):
        """Test non-convex QP which xpress_direct should prove infeasible."""
        m = pyo.ConcreteModel()
        m.x1 = pyo.Var()
        m.x2 = pyo.Var()
        m.x3 = pyo.Var()

        m.obj = pyo.Objective(rule=lambda m: 2 * m.x1 + m.x2 + m.x3, sense=pyo.minimize)
        m.equ1a = pyo.Constraint(rule=lambda m: m.x1 + m.x2 + m.x3 == 1)
        m.equ1b = pyo.Constraint(rule=lambda m: m.x1 + m.x2 + m.x3 == -1)
        m.cone = pyo.Constraint(rule=lambda m: m.x2 * m.x2 + m.x3 * m.x3 <= m.x1 * m.x1)
        m.equ2 = pyo.Constraint(rule=lambda m: m.x1 >= 0)

        opt = pyo.SolverFactory('xpress_direct')
        opt.options['XSLP_SOLVER'] = 0

        results = opt.solve(m)
        self.assertEqual(results.solver.status, SolverStatus.ok)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.infeasible
        )


@unittest.pytest.mark.solver("xpress_persistent")
class TestXpressPersistentMock(unittest.TestCase):
    def test_available(self):
        class mock_xpress:
            def __init__(self, importable, initable):
                self._initable = initable
                xpd.xpress_available = importable

            def log_import_warning(self, logger):
                logging.getLogger(logger).warning("import warning")

            def init(self):
                if not self._initable:
                    raise RuntimeError("init failed")

            def free(self):
                pass

        orig = xpd.xpress, xpd.xpress_available
        try:
            _xpress_persistent = XpressPersistent
            xpd.xpress = mock_xpress(True, True)
            with LoggingIntercept() as LOG:
                self.assertTrue(XpressPersistent().available(True))
                self.assertTrue(XpressPersistent().available(False))
            self.assertEqual(LOG.getvalue(), "")

            xpd.xpress = mock_xpress(False, False)
            with LoggingIntercept() as LOG:
                self.assertFalse(XpressPersistent().available(False))
            self.assertEqual(LOG.getvalue(), "")
            with LoggingIntercept() as LOG:
                with self.assertRaisesRegex(
                    xpd.ApplicationError,
                    "No Python bindings available for .*XpressPersistent.* "
                    "solver plugin",
                ):
                    XpressPersistent().available(True)
            self.assertEqual(LOG.getvalue(), "import warning\n")

            xpd.xpress = mock_xpress(True, False)
            with LoggingIntercept() as LOG:
                self.assertFalse(XpressPersistent().available(False))
            self.assertEqual(LOG.getvalue(), "")
            with LoggingIntercept() as LOG:
                with self.assertRaisesRegex(RuntimeError, "init failed"):
                    XpressPersistent().available(True)
            self.assertEqual(LOG.getvalue(), "")
        finally:
            xpd.xpress, xpd.xpress_available = orig



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Pyomo plugin for interacting with NEOS solvers
#

# Static documentation for NEOS solvers
doc = {
    'baron': 'Deterministic global MINLP solver',
    'bonmin': 'Deterministic local MINLP solver',
    'cbc': 'MILP solver',
    'conopt': 'Feasible path NLP solver',
    'couenne': 'Deterministic global MINLP solver',
    'cplex': 'MILP solver',
    'filmint': 'Deterministic local MINLP solver',
    'filter': 'SQP NLP solver',
    'ipopt': 'Interior point NLP solver',
    'knitro': 'Convex NLP/MINLP solver',
    'l-bfgs-b': 'Bound-constrained NLP solver',
    'lancelot': 'Augmented Lagrangian NLP solver',
    'lgo': 'Lipschitz-Continuous Global Optimizer',
    'loqo': 'Interior point NLP solver',
    'minlp': 'Deterministic local MINLP solver',
    'minos': 'NLP solver',
    'minto': 'MILP solver',
    'mosek': 'Interior point NLP solver',
    # 'octeract': 'Deterministic global MINLP solver',
    'ooqp': 'Convex QP solver',
    'path': 'Nonlinear MCP solver',
    'raposa': 'A Global Solver for Polynomial Programming Problems',
    'snopt': 'SQP NLP solver',
}



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for SymbolMap
#

import pyomo.common.unittest as unittest
from pyomo.environ import (
    ConcreteModel,
    Set,
    Var,
    Objective,
    Constraint,
    Block,
    SymbolMap,
    TextLabeler,
)
from pyomo.core.base.symbol_map import symbol_map_from_instance


class Test(unittest.TestCase):
    def setUp(self):
        #
        # Create model instance
        #
        model = ConcreteModel()
        model.A = Set(initialize=[1, 2, 3])
        model.x = Var()
        model.y = Var(model.A, dense=True)
        model.o1 = Objective(expr=model.x)

        def o2_(model, i):
            return model.x

        model.o2 = Objective(model.A, rule=o2_)
        model.c1 = Constraint(expr=model.x >= 1)

        def c2_(model, i):
            if i == 1:
                return model.x <= 2
            elif i == 2:
                return (3, model.x, 4)
            else:
                return model.x == 5

        model.c2 = Constraint(model.A, rule=c2_)
        model.b = Block()
        model.b.x = Var()
        model.b.y = Var(model.A, dense=True)
        self.instance = model

    def tearDown(self):
        self.instance = None

    def test_add(self):
        smap = SymbolMap()
        smap.addSymbol(self.instance.x, "x")
        smap.addSymbol(self.instance.y[1], "y[1]")
        self.assertEqual(set(smap.bySymbol.keys()), set(['x', 'y[1]']))

    def test_adds(self):
        smap = SymbolMap()
        labeler = TextLabeler()
        smap.addSymbols(
            (obj, labeler(obj)) for obj in self.instance.component_data_objects(Var)
        )
        self.assertEqual(
            set(smap.bySymbol.keys()),
            set(['x', 'y(1)', 'y(2)', 'y(3)', 'b_x', 'b_y(1)', 'b_y(2)', 'b_y(3)']),
        )

    def test_create(self):
        smap = SymbolMap()
        labeler = TextLabeler()
        smap.createSymbol(self.instance.x, labeler)
        smap.createSymbol(self.instance.y[1], labeler)
        self.assertEqual(set(smap.bySymbol.keys()), set(['x', 'y(1)']))

    def test_creates(self):
        smap = SymbolMap()
        labeler = TextLabeler()
        smap.createSymbols(self.instance.component_data_objects(Var), labeler)
        self.assertEqual(
            set(smap.bySymbol.keys()),
            set(['x', 'y(1)', 'y(2)', 'y(3)', 'b_x', 'b_y(1)', 'b_y(2)', 'b_y(3)']),
        )

    def test_get(self):
        smap = SymbolMap()
        labeler = TextLabeler()
        self.assertEqual('x', smap.getSymbol(self.instance.x, labeler))
        self.assertEqual('y(1)', smap.getSymbol(self.instance.y[1], labeler))
        self.assertEqual(set(smap.bySymbol.keys()), set(['x', 'y(1)']))
        self.assertEqual('x', smap.getSymbol(self.instance.x, labeler))

    def test_alias_and_getObject(self):
        smap = SymbolMap()
        smap.addSymbol(self.instance.x, 'x')
        smap.alias(self.instance.x, 'X')
        self.assertEqual(set(smap.bySymbol.keys()), set(['x']))
        self.assertEqual(set(smap.aliases.keys()), set(['X']))
        self.assertEqual(id(smap.getObject('x')), id(self.instance.x))
        self.assertEqual(id(smap.getObject('X')), id(self.instance.x))

    def test_from_instance(self):
        smap = symbol_map_from_instance(self.instance)
        self.assertEqual(
            set(smap.bySymbol.keys()),
            set(
                [
                    'x',
                    'y(1)',
                    'y(2)',
                    'y(3)',
                    'b_x',
                    'b_y(1)',
                    'b_y(2)',
                    'b_y(3)',
                    'o1',
                    'o2(1)',
                    'o2(2)',
                    'o2(3)',
                    'c1',
                    'c2(1)',
                    'c2(2)',
                    'c2(3)',
                ]
            ),
        )
        self.assertEqual(
            set(smap.aliases.keys()),
            set(
                [
                    'c_e_c2(3)_',
                    '__default_objective__',
                    'c_u_c2(1)_',
                    'c_l_c1_',
                    'r_u_c2(2)_',
                    'r_l_c2(2)_',
                ]
            ),
        )

    def test_error1(self):
        smap = SymbolMap()
        labeler = TextLabeler()
        self.assertEqual('x', smap.getSymbol(self.instance.x, labeler))

        class FOO:
            def __call__(self, *args):
                return 'x'

        labeler = FOO()
        try:
            self.assertEqual('x', smap.getSymbol(self.instance.y[1], labeler))
            self.fail("Expected RuntimeError")
        except RuntimeError:
            pass

    def test_error2(self):
        smap = SymbolMap()
        smap.addSymbol(self.instance.x, 'x')
        smap.alias(self.instance.x, 'X')
        self.assertEqual(id(smap.getObject('x')), id(self.instance.x))
        self.assertEqual(id(smap.getObject('X')), id(self.instance.x))
        self.assertEqual(id(smap.getObject('y')), id(SymbolMap.UnknownSymbol))


if __name__ == "__main__":
    unittest.main()
